
# Copyright (c) 2023 Watsen Networks.  All Rights Reserved.

_v='config-true-seq-nodes'
_u='config-false-prefixes'
_t='table-name-map'
_s='table-keys'
_r='app_ns'
_q='sztpd_meta'
_p='SELECT schema_name FROM information_schema.schemata;'
_o='postgresql'
_n='Cannot delete: '
_m='[^/]*=[^/]*'
_l='[^:]*:'
_k='yang-library'
_j='" does not exist.'
_i=':tenants/tenant'
_h='global-root'
_g='db_ver'
_f='sqlite'
_e=':memory:'
_d='key'
_c='first'
_b='Parent node ('
_a='.singletons'
_Z='mysql'
_Y='" already exists.'
_X='Node "'
_W='sort-by'
_V='.*/'
_U='last'
_T=') does not exist.'
_S='_idx'
_R='_key'
_Q='after'
_P='before'
_O='insert'
_N='ssl'
_M='row_id'
_L=':'
_K='point'
_J='pid'
_I='jsob'
_H='obu_idx'
_G='singletons'
_F='=[^/]*'
_E=True
_D='='
_C=False
_B='/'
_A=None
import os,re,sys,json,base64,pickle,yangson,binascii,pkg_resources,sqlalchemy as sa
from enum import IntFlag
from urllib.parse import quote
from urllib.parse import unquote
from sqlalchemy.sql import and_
from dataclasses import dataclass
from sqlalchemy.sql import bindparam
from sqlalchemy.schema import CreateTable
from .  import db_utils
'\nSQL-based DAL\n\nFirstly, note that the entire persistent store is an facade for a arbitrary hierarchical N-ary tree\nof data, accessed by HTTP operations on URLs.\n\nInternally, this DAL maps the tree to tables as follows:\n\n    1) there is one table called \'singletons\' which contains things that are known to only occur\n       once: metadata (db_ver, key-test, table-defs.), the global root of the data tree (and all\n       descendants up nodes up to the point they\'ve been fanned out to a distinct database table).\n       Each row of this table is its own distinct thing of data, unrelated to any other row in\n       the table.  There are just two columns: \'name\' and \'jsob\' (json object).\n\n       this JSOB has the format:\n\n            {\n                "namespace:node-1" : {-or-[ ... ]-or-},\n                "namespace:node-1" : {-or-[ ... ]-or-},\n                "namespace:node-1" : {-or-[ ... ]-or-}\n                ...\n            }\n\n    2) there is a database table for each node in the YANG model of type \'list\'.  That is, the\n       list is a table and each entry in the list is a row in the table.\n\n       A SQL table is created with its name being the schema_path of the \'list\' node.  This \n       results in some surprisingly long table names, but leads to many benefits in the code.\n\n       Each \'list\' table contains at least one column: \'id\' (a sequential row identifier).\n       \n       If the table is a descendent of another table, then the table will contain another column:\n       \'parent-id\' (the \'id\' of the row in the parent table to which this row is a child).  Note\n       that this construct can support arbitrarily nested tables.\n\n       If the list node is "config true", then the table will contain two more columns: \'key\'\n       and \'jsob\'.   The \'key\' column contains the list\'s key value as a string, allowing for\n       fast lookups.  Typically the table\'s primary key would be the 2-tuple \'parent-id + key\',\n       so that uniqueness per list instance (matching YANG semantics), but if the key node\n       contains the \'globally-unique\' extension contains, then the primary key only includes\n       the \'key\' column (excluding the \'parent-id\' column).  The \'jsob\' (JSON Object) contains\n       a JSON document containing all the node\'s data, including the \'key\' field, but excluding\n       any data belonging to a descendent list/table.\n\n       If the list node is "config false", then the table will contain a column for each descendent\n       node of the \'list\' node.  Each column is appropriately typed (string, integer, datetime,\n       json, etc.), and all simple-typed columns are indexed.  Currently, the "globally-unqiue"\n       extension on the \'key\' node is not supported.\n\n       These JSOBs have the format:  (note that there may be more than one top-level \'dict\')\n            {\n                "namespace:node-1" : { ... },   # prefix always exists, as the DS-root is namespaceless\n                "namespace:node-2" : { ... },   # ditto\n                "namespace:node-3" : { ... }    # ditto\n                ...\n            } \n                         FIXME: how to support top-level lists?  (do they not show up at all in singleton\'s JSOB?)\n\n       For instance, for /wn-sztpd-0:transport/listen/endpoint (this is a \'list\'), each list-entry\n       (i.e., table row) is like:   (note that there may be only one top-level \'dict\')\n\n            {\n                "endpoint" : { ... }\n            }\n\n       In another example, for /admin-accounts/admin-account (this is a \'list\'), each list-entry\n       (i.e., table row) is like:   (note that there may be only one top-level \'dict\')\n\n            {\n                "admin-account" : { ... }\n            }\n\n    \n\n\n    === IGNORE FOR NOW ===\n    3) there is a database table called \'ref-stats\' that contains a row for each node in\n       the YANG model that has been tagged with the special "ref-stats" extension.  This row\n       counts the number of \'leafref\' instances to it or to any of its descendant nodes, as\n       well as tracks when references are added/removed (a single timestamp, not a log).  The\n       tracking of ref-counts to descendant nodes is rather common, as leafrefs are many times\n       to a list\'s "key" field(s) (e.g., a descendant node called "name").  There may exist\n       nested \'ref-stats\' node (e.g., the "certificate" list inside the "assymetric-key" list),\n       in such case it is expected that the ancestor ref-count includes the sum of all the\n       descendant counts plus any additional references to it.  Be aware that while it is most\n       common for a list element to have a \'ref-stats\', it is entirely reasonable for a scaler\n       element to be ref-counted too.\n\n       Side note: the ref counts and timestamps are primarily used by the system itself (not\n       end users) to detect when certain nodes are no longer referenced (i.e., orphaned) and\n       hence can be auto-deleted.  That said, exposing said values to users is goodness too.\n\n\n    FUTURE THOUGHTS\n    ---------------\n    A super-clever implementation would have no \'db-table\' annotations and would simply\n    initialize with just monster-sized top-level jsobs - and then use runtime statistics\n    to evaluate how the database is typically used and dyanamically reset tables as needed.\n\n    But that sounds hard (and perhaps somewhat stupid), perhaps a better idea is to analyze\n    the schema more to find logical break points (e.g., lists with large subtrees, lists\n    with *lots* of elements (i.e., config false), anything that collects ref-stats, etc.?\n\n\n\n\n\n\n\n'
known_text=b'a secret message'
jsob_type=sa.JSON
class ContentType(IntFlag):CONFIG_TRUE=1;CONFIG_FALSE=2;CONFIG_ANY=3
@dataclass
class DatabasePath:data_path:str=_A;schema_path:str=_A;jsob_data_path:str=_A;jsob_schema_path:str=_A;table_name:str=_A;row_id:int=_A;inside_path:str=_A;path_segments:list=_A;jsob:dict=_A;node_ptr:dict=_A;prev_ptr:dict=_A
class AuthenticationFailed(Exception):0
class NodeAlreadyExists(Exception):0
class NodeNotFound(Exception):0
class ParentNodeNotFound(Exception):0
class TooManyNodesFound(Exception):0
class InvalidResourceTarget(Exception):0
class CreateCallbackFailed(Exception):0
class ChangeCallbackFailed(Exception):0
class CreateOrChangeCallbackFailed(Exception):0
class DataAccessLayer:
	'\n    This class contains a five "handle" routines having the signature:\n\n        async def handle_<operation>_<context>_request(self, request)\n\n    Where:\n\n        <operation> is an HTTP operations: get, post, put, and delete.\n                    [FIXME: add support for \'patch\' too!]\n\n        <context> is one of "opstate" or "config"\n                  [Note: the "opstate" context only supports "get"]\n\n    As can be seen from the signature, each of these routines are "async"\n    and take as a parameter a "request" object, the aiohttp-based HTTP\n    request from the RestconfServer layer.\n\n    These routines are only called from higher-level layers are, in\n    particular, as never recursive.  As such, each routine opens a\n    database connection (via a context manager) used by all the DB\n    calls throughout its processing.  Furthermore, for routines\n    that *write* data to the database, a transaction is started\n    using this connection, which is subsequently committed or\n    rollbacked when the routine exits, respectively whether\n    successful or not (e.g., an exception is raised).\n\n    Note that, due to the nature of working with arbitrary trees, these\n    top-level routines often times call on a supporting recursive method.\n    '
	def __init__(A,db_url,cacert_param=_A,cert_param=_A,key_param=_A,yl_obj=_A,app_ns=_A,opaque=_A):
		'\n        db_url:        a SqlAlchemy URL string\n        cacert_param:  path to PEM file containing a list of X.509 certificate\n        cert_param:    path to PEM file containing the client certificate (FIXME: must be a single cert, MySQL-only?)\n        key_param:     path to PEM file containing a key for the client certificate\n        yl_obj:        a Python object containing the yang-library instance used\n                       to initialize a new database with.\n        opaque:        an app-specific python object that can be persisted when\n                       the database is created and returned thereafter.\n        ';F=yl_obj;E=key_param;D=cert_param;C=cacert_param;B=db_url;A.app_ns=_A;A.engine=_A;A.metadata=_A;A.leafrefs=_A;A.referers=_A;A.ref_stat_collectors=_A;A.global_root_id=_A;A.table_keys=_A;A.schema_path_to_real_table_name=_A;A.config_false_prefixes=_A;A.post_dal_callbacks=_A;A.config_true_obu_seq_nodes=_A
		if F is _A:A._init(B,C,D,E)
		else:A.app_ns=app_ns;A._create(B,C,D,E,F,opaque)
		assert A.app_ns!=_A;assert A.engine!=_A;assert A.metadata!=_A;assert A.leafrefs!=_A;assert A.referers!=_A;assert A.ref_stat_collectors!=_A;assert A.global_root_id!=_A;assert A.table_keys!=_A;assert A.schema_path_to_real_table_name!=_A;assert A.config_false_prefixes!=_A;assert A.config_true_obu_seq_nodes!=_A;A.len_app_ns=len(A.app_ns)
	async def num_elements_in_list(A,data_path):
		" Returns the number of elements in the list specified by the passed data-path.\n            The data-path must be for the 'list' node (not its parent node).\n    \n            Note: this is a very special routine that is currently only used to detect if\n                  any admins have been configured, so as to enable/disable auth testing!\n        ";B=data_path;C=re.sub(_F,'',B)
		if C!=B:assert NotImplementedError("Nested listed arren't supported yet.")
		D=A.schema_path_to_real_table_name[C];E=A.metadata.tables[D]
		with A.engine.connect()as F:G=F.execute(sa.select(sa.func.count()).select_from(E));H=G.first()[0];return H
	async def get_tenant_name_for_admin(A,email_address):
		" Returns tenant's name or None (for root)\n            raise NodeNotFound if admin doesn't exist.\n            Only works for mode-'x'.  Calling code must ensure this.\n        ";I='email-address';G=email_address;B=A.schema_path_to_real_table_name[_B+A.app_ns+':tenants/tenant/admin-accounts/admin-account'];C=A.metadata.tables[B]
		with A.engine.connect()as E:
			D=E.execute(sa.select(C.c.pid).where(C.c[I]==G));F=D.first()
			if F!=_A:J=F[0];B=A.schema_path_to_real_table_name[_B+A.app_ns+_i];H=A.metadata.tables[B];D=E.execute(sa.select(H.c.name).where(H.c.row_id==J));K=D.first()[0];return K
		B=A.schema_path_to_real_table_name[_B+A.app_ns+':admin-accounts/admin-account'];C=A.metadata.tables[B]
		with A.engine.connect()as E:
			D=E.execute(sa.select(C.c.pid).where(C.c[I]==G));F=D.first()
			if F!=_A:return _A
		raise NodeNotFound('Admin "'+G+_j)
	async def get_tenant_name_for_global_key(A,table_name,k):
		" Returns tenant's name for the record in table_name having matching key 'k'\n            raise NodeNotFound if k doesn't exist.  Only works for mode-'x'.\n            Calling code must ensure this.\n        ";B=table_name;C=A.schema_path_to_real_table_name[B];E=A.metadata.tables[C]
		with A.engine.connect()as F:
			D=F.execute(sa.select(E.c.pid).where(getattr(E.c,A.table_keys[B])==k));G=D.first()
			if G==_A:raise NodeNotFound('key "'+k+'" in table "'+B+_j)
			I=G[0];C=A.schema_path_to_real_table_name[_B+A.app_ns+_i];H=A.metadata.tables[C];D=F.execute(sa.select(H.c.name).where(H.c.row_id==I));J=D.first()[0];return J
	async def handle_get_opstate_request(B,data_path,query_dict={}):
		'\n        Pretty much a pass-thru to _handle_get_data_request()\n        ';C=query_dict;A=data_path;assert A!='';assert not(A!=_B and A[-1]==_B)
		if A=='/ietf-yang-library:yang-library':
			F=B.schema_path_to_real_table_name[_G];D=B.metadata.tables[F]
			with B.engine.connect()as G:H=G.execute(sa.select(D.c.jsob).where(D.c.name==_k));return H.first()[0]
		E=re.sub(_F,'',A)
		if 0:return await B._handle_get_data_request(A,E,ContentType.CONFIG_FALSE,C)
		else:return await B._handle_get_data_request(A,E,ContentType.CONFIG_ANY,C)
	async def handle_get_config_request(B,data_path,query_dict={}):
		'\n        Fetches data using ContentType.CONFIG_TRUE, but that only prevents data from\n        opstate-tables (which could be huge) from being included. However, the result\n        still contains non-list opstate data, which needs to be cherry-picked out...\n        ';E=data_path;A=re.sub(_F,'',E);F=await B._handle_get_data_request(E,A,ContentType.CONFIG_TRUE,query_dict);C=re.sub(_V,'',A)
		if not C.startswith(B.app_ns+_L):C=B.app_ns+_L+C
		for D in B.config_false_prefixes:
			if D.startswith(A):
				if A==_B:G=D[1:]
				else:G=D.replace(A,C,1)
				I=_A;J=F
				def H(prev_ptr,curr_ptr,remainder_path):
					D=prev_ptr;B=remainder_path;A=curr_ptr;E=B.split(_B)
					for C in E:
						if type(A)==list:
							for F in A:H(A,F,B)
							return
						elif C in A:D=A;A=A[C];B=B.replace(C+_B,'',1)
						else:return
					D.pop(C)
				H(I,J,G)
			else:0
		return F
	async def _handle_get_data_request(A,data_path,schema_path,content_type,query_dict):
		'\n        Recursion strategy:\n          - extract node from appropriate DB-table to create JSOB\n          - then recursively descend jsob/sub-tables, adding into JSOB as it goes\n          - for each JSON, prune out CONFIG FALSE nodes if not wanting them\n          - any failure due to object not existing only occurs at top\n        ';N='Node (';I=content_type;G=schema_path;F=data_path
		if I is ContentType.CONFIG_TRUE and any((G.startswith(B)for B in A.config_false_prefixes)):raise NodeNotFound(N+F+_T)
		with A.engine.connect()as J:
			B=A._get_dbpath_for_data_path(F,I,J)
			if B==_A:raise NodeNotFound(N+F+_T)
			if F!=_B and B.schema_path in A.schema_path_to_real_table_name:
				K,O=F.rsplit(_B,1)
				if _D not in O:
					if K=='':L=_B
					else:L=re.sub(_F,'',K)
					E={};await A._recursively_attempt_to_get_data_from_subtable(B.schema_path,B.row_id,L,B.jsob_data_path,E,I,query_dict,J)
					if E!={}:
						C=next(iter(E))
						if not C.startswith(A.app_ns+_L):E[A.app_ns+_L+C]=E[C];E.pop(C)
					else:
						C=G.rsplit(_B,1)[1]
						if not C.startswith(A.app_ns+_L):C=A.app_ns+_L+C
						E[C]=[]
					return E
			D=re.sub(_V,'',G);P=D
			if D=='':H=B.node_ptr
			else:
				if not D.startswith(A.app_ns+_L):D=A.app_ns+_L+D
				H={}
				if B.table_name!=_G and P==B.inside_path:H[D]=[];H[D].append(B.node_ptr);B.node_ptr=H[D][0]
				else:H[D]=B.node_ptr
			Q=A._get_list_of_direct_subtables_for_schema_path(G)
			for M in Q:R=F+M[len(G):];await A._recursively_attempt_to_get_data_from_subtable(M,B.row_id,G,R,B.node_ptr,I,{},J)
		return H
	async def _recursively_attempt_to_get_data_from_subtable(D,subtable_name,pid,jsob_schema_path,jsob_data_path,jsob_iter,content_type,query_dict,conn):
		'\n        This routine supports the top-level _handle_get_data_request() routine.\n\n        subtable_name: the name of the sub-table to check if data from\n        pid: the parent table\'s "id"\n        jsob_schema_path: an arbitrary schema path (mapped from HTTP client\'s data_path).  Must begin with \'/\' (commonly "/wn-sztpd-?:")\n        jsob_data_path: an arbitrary data_path\n        jsob_iter: a pointer into a DOM being constructed at the spot pointed to by \'jsob_schema_path\'.  It\'s top-node is NOT the path\'s tail segment.\n        conn: the database connection to use (no transaction for \'get\')\n\n        NOTE TO FUTURE SELF:\n          - a "keys-only" query parameter will be added to enable clients to obtain more manageable responses\n             - "off" by default (backwards compatible add)\n             - for "config true" tables, return just the keys\n             - for "config false" tables, return either an empty list, metadata w/ a count, or a remediation msg\n                - metadata?: https://tools.ietf.org/html/rfc7952#section-5.2.2\n        ';K=query_dict;J=content_type;I=jsob_schema_path;C=subtable_name
		if not ContentType.CONFIG_FALSE in J:
			if any((C.startswith(A)for A in D.config_false_prefixes)):return
			else:0
		else:0
		if C in D.config_true_obu_seq_nodes:assert _W not in K;K[_W]=_H
		S=D._find_rows_in_table_having_pid(C,pid,K,conn);L=S.fetchall()
		if len(L)==0:return
		B=jsob_iter
		if I==_B:O=C[1:]
		else:assert I.startswith(_B+D.app_ns);O=C.replace(I+_B,'')
		E=re.sub(_V,'',C);G=re.sub(E+'$','',O)
		if G!='':
			G=G[:-1];T=G.split(_B)
			for M in T:
				try:B=B[M]
				except KeyError as X:assert ContentType.CONFIG_FALSE in J;B[M]={};B=B[M]
		U=D.schema_path_to_real_table_name[C];P=D.metadata.tables[U]
		if any((C.startswith(A)for A in D.config_false_prefixes)):
			B[E]=[]
			for F in L:
				H={}
				for A in P.c:
					if A.name!=_M and A.name!=_J:
						if type(A.type)is sa.sql.sqltypes.DateTime:H[A.name]=F[A.name].strftime('%Y-%m-%dT%H:%M:%SZ')
						elif type(A.type)is sa.sql.sqltypes.JSON or type(A.type)is sa.sql.sqltypes.PickleType:
							if F[A.name]is not _A and not(type(F[A.name])is list and len(F[A.name])==0):H[A.name]=F[A.name]
						elif F[A.name]is not _A:H[A.name]=F[A.name]
				B[E].append(H)
		else:
			Q=0
			for N in L:
				R=N[_I];E=next(iter(R))
				if E not in B:B[E]=[]
				B[E].append(R.pop(E));V=D._get_list_of_direct_subtables_for_schema_path(C)
				for W in V:await D._recursively_attempt_to_get_data_from_subtable(W,N[_M],C,jsob_data_path+_D+N[P.c[2].name],B[E][Q],J,{},conn)
				Q+=1
	async def handle_post_config_request(A,data_path,query_dict,request_body,create_callbacks,change_callbacks,opaque):
		'\n        Parameters:\n          self: dal\n          data_path: the data_path of the node to create\n          request_body: the data definition\n          create_callbacks: a dict, keyed by schema paths, mapping to a list of callbacks\n          change_callbacks: a dict, keyed by schema paths, mapping to a list of callbacks (only called for this object)\n          opaque: an opaque object that is passed into the callback\n\n        Recursion strategy:\n          - recursively descend into jsob/sub-tables (pro-tip: recursive calls only occur at DB-table boundaries)\n          - on exit, extract from JSOB and insert into DB-table\n          - note that any failure due to object already existing will only ever occur at top\n        ';E=request_body;B=data_path;assert B!='';assert not(B!=_B and B[-1]==_B);assert'?'not in B;assert type(E)==dict
		with A.engine.begin()as C:
			F=A._get_dbpath_for_data_path(B,ContentType.CONFIG_TRUE,C)
			if F==_A:raise ParentNodeNotFound(_b+B+_T)
			await A._handle_post_config_request(F,query_dict,E,create_callbacks,change_callbacks,opaque,C)
			if A.post_dal_callbacks is not _A:
				for D in A.post_dal_callbacks:
					try:await D[0](D[1],C,D[2])
					except Exception as G:A.post_dal_callbacks=_A;raise G
				A.post_dal_callbacks=_A
	async def _handle_post_config_request(B,parent_dbpath,query_dict,request_body,create_callbacks,change_callbacks,opaque,conn):
		"\n        Similar to handle_post_config_request() but:\n          - no input validation\n          - no starting of a new transaction\n          - parent node dbpath opened (caller MUST save it out)\n\n        This routine is used by:\n          - handle_post_config_request()   [for obvious reasons]\n          - handle_put_config_request()    [for when PUT implcitily creates (i.e., acts like a POST)]\n          - _handle_put_config_request()   [for when PUT creates internal nodes within itself]\n\n        Parameters:\n          - self: DAL\n          - parent_dbpath: the 'dbpath' for the parent node\n          - request_body: the descendent node to create\n          - create_callbacks: dict containing 'create' callbacks\n          - opaque: arbitrary type passed into callbacks\n          - conn: current database connection\n        ";a=change_callbacks;V=opaque;U=create_callbacks;K=conn;J=query_dict;F=request_body;A=parent_dbpath;O=A.data_path;b=re.sub(_F,'',O);E=next(iter(F))
		if O==_B:C=E;L=_B+C
		else:C=re.sub(_l,'',E);L=O+_B+C
		M=re.sub(_F,'',L);H=B._get_table_name_for_schema_path(M);assert H!=_A
		if H==A.table_name:
			if C in A.node_ptr:
				if type(A.node_ptr[C])!=list:raise NodeAlreadyExists(_X+C+_Y)
				assert len(F[E])==1
				if F[E][0]in A.node_ptr[C]:raise NodeAlreadyExists(_X+F[E][0]+_Y)
				if M not in B.config_true_obu_seq_nodes:A.node_ptr[C].append(F[E][0])
				else:
					G=_U
					if _O in J:G=J[_O]
					if G==_c:assert _K not in J;A.node_ptr[C].insert(0,F[E][0])
					elif G==_U:assert _K not in J;A.node_ptr[C].append(F[E][0])
					else:
						assert G in(_P,_Q);assert _K in J;P=J[_K].rsplit(_D,1)[1];P=unquote(P);D=A.node_ptr[C].index(P)
						if G==_P:0
						elif G==_Q:D=D+1
						A.node_ptr[C].insert(D,F[E][0])
				if M in U:
					for Y in U[M]:await Y(L+_D+F[E][0],A.jsob,A.jsob_data_path,V)
			else:A.node_ptr[C]=F.pop(E);c=await B._recursively_post_subtable_data(A.row_id,L,A.node_ptr[C],A.jsob,A.jsob_data_path,U,V,K)
			B._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,K)
		else:
			assert M in B.table_keys
			if C not in A.node_ptr:A.node_ptr[C]=[];B._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,K)
			Q=F[E][0][B.table_keys[H]]
			if M in B.config_true_obu_seq_nodes:I=B._get_obu_idx_list_for_list_path(L,K)
			W={};assert len(F[E])==1;W[C]=F[E][0];R={};R[_J]=A.row_id;R[B.table_keys[H]]=Q
			if M in B.config_true_obu_seq_nodes:R[_H]=-1
			R[_I]={};S=B.schema_path_to_real_table_name[H];N=B.metadata.tables[S]
			try:Z=K.execute(N.insert().values(**R))
			except sa.exc.IntegrityError:raise NodeAlreadyExists(_X+C+_Y)
			L+=_D+str(Q);c=await B._recursively_post_subtable_data(Z.inserted_primary_key[0],L,W[C],W,L,U,V,K);S=B.schema_path_to_real_table_name[H];N=B.metadata.tables[S];K.execute(N.update().where(N.c.row_id==Z.inserted_primary_key[0]).values(jsob=W))
			if M in B.config_true_obu_seq_nodes:
				if O==_B:d=_B+C
				else:d=O+_B+C
				G=_U
				if _O in J:G=J[_O]
				if G==_c:D=0;T=[(B.table_keys[H],Q),(_H,D)];I.insert(0,T)
				elif G==_U:D=len(I);T=[(B.table_keys[H],Q),(_H,D)];I.append(T)
				else:
					assert G in(_P,_Q);P=J[_K].rsplit(_D,1)[1];D=0
					for X in I:
						if X[0][1]==P:break
						D+=1
					if G==_P:0
					elif G==_Q:D=D+1
					T=[(B.table_keys[H],Q),(_H,D)];I.insert(D,T)
				for X in range(D+1,len(I)):I[X][1]=_H,I[X][1][1]+1
				S=B.schema_path_to_real_table_name[H];N=B.metadata.tables[S];e=B.table_keys[H];f=N.update().where(and_(N.c.pid==A.row_id,getattr(N.c,e)==bindparam(_R))).values(obu_idx=bindparam(_S));Z=K.execute(f,[{_R:A[0][1],_S:A[1][1]}for A in I[D:len(I)]])
			else:0
		g=re.sub(_F,'',O)
		if g in a:
			for Y in a[b]:await Y(O,A.jsob,A.jsob_data_path,V)
	async def _recursively_post_subtable_data(B,pid,data_path,req_body_iter,jsob,jsob_data_path,create_callbacks,opaque,conn):
		"\n        Prune incoming req_body_iter, extracting subtable data out.\n        Calling routine MUST handle saving out any parent data.\n\n        Extract and persist req_body_iter into subtables\n          - will open and save jsobs on table boundaries\n          - req_body_iter is pruned in-place (i.e., the parts put into subtables are removed)\n          - jsob: the parent jsob or None, if this call creates a new list entry (i.e., on a list boundary)\n          - jsob_data_path: the data_path for the parent jsob or None, if this call creates a new list entry (i.e., on a list boundary)\n          - data_path tracking isn't needed for persisting, but is needed for inline callbacks...\n\n        Parameters:\n          - pid: parent table's row-id\n          - data_path: data_path in the request_body tree \n          - req_body_iter: pointer to this place in the request_body tree\n          - (OLD) prev_req_body_iter: pointer to 'req_body_iter' parent node (for inline callbacks)\n          - jsob: pointer to 'jsob' this node is inside (for inline callbacks)\n          - jsob_data_path: the 'data_path' for the 'jsob' passed\n          - create_callbacks: a dictionary (keyed by schema_path) of callbacks to execute when matching nodes are created\n          - opaque: an arbitrary Python object passed into the create callbacks.\n          - conn: the currently opened sqlalchemy database connection\n\n        Returns True if a list, False otherwise (enables caller to trim dict as needed)\n        ";P=pid;M=jsob_data_path;L=jsob;I=conn;H=opaque;E=create_callbacks;D=data_path;A=req_body_iter;C=re.sub(_F,'',D)
		if type(A)==dict:
			'\n            simple recursion, but need to handle case where child is a list\n            '
			if C in E:
				for Q in E[C]:await Q(D,L,M,H)
			for N in A.copy():
				if N.startswith(B.app_ns):F=D+_B+N[B.len_app_ns+1:]
				else:F=D+_B+N
				R=await B._recursively_post_subtable_data(P,F,A[N],L,M,E,H,I)
		elif type(A)==list:
			if C in B.table_keys:
				if C in B.config_true_obu_seq_nodes:S=0
				while A:
					J=A.pop(0);assert type(J)==dict;V=re.sub(_V,'',C);O={};O[V]=J;K=B.table_keys[C];G={};G[_J]=P;G[K]=J[K]
					if C in B.config_true_obu_seq_nodes:G[_H]=S;S+=1
					G[_I]=O;W=B.schema_path_to_real_table_name[C];X=B.metadata.tables[W]
					try:T=I.execute(X.insert().values(**G))
					except sa.exc.IntegrityError as Y:raise NodeAlreadyExists(_X+K+'" with value "'+G[K]+_Y)
					F=D+_D+str(J[K]);R=await B._recursively_post_subtable_data(T.inserted_primary_key[0],F,J,O,F,E,H,I);B._update_jsob_for_row_id_in_table(C,T.inserted_primary_key[0],O,I)
				assert type(A)==list;assert len(A)==0
			else:
				assert type(A)==list
				if not(len(A)==1 and A[0]==_A):
					for U in A:F=D+_D+str(U);R=await B._recursively_post_subtable_data(P,F,A[A.index(U)],L,M,E,H,I)
				else:0
		elif C in E:
			for Q in E[C]:await Q(D,L,M,H)
		if C in B.table_keys:return _E
		return _C
	async def handle_post_opstate_request(A,data_path,request_body):
		"\n        This routine is only used to support internal application logic.\n          - it does NOT support RESTCONF's POST!\n\n        Parameters:\n            data_path: the table name.\n            request_body: the jsob to insert\n\n        This routine is only handles one use case: inserting a new row into an \n        'opstate' table (e.g., audit log, alarms).\n       \n        This routine is NOT expected to insert an opstate node into into a heirarchy\n        of (config) nodes.\n        \n        Note that the parent config-false container is assumed to already exist.\n        That is, even before any entries have been inserted into the opstate\n        table, a GET (opstate) on the parent container should SUCCEED with an\n        empty response.  As such, this routine does NOT test for or auto-create\n        the parent container.\n\n        Recursion strategy: None - all and any sub-lists (or whatever) are\n        stored as JSOBs.\n        ";M='Unrecognized resource schema path: ';I=request_body;E=data_path;B=re.sub(_F,'',E);F=next(iter(I))
		if B==_B:G=F;C=_B+G
		else:G=re.sub(_l,'',F);C=B+_B+G
		J=A._get_table_name_for_schema_path(C)
		if J!=C:raise NodeNotFound(M+C)
		D=I[F];K=re.findall(_m,E)
		if len(K)==0:D[_J]=A.global_root_id
		else:
			L=A._get_table_name_for_schema_path(B)
			if L==_A:raise ParentNodeNotFound(M+B)
			N=K[-1].split(_D)
			with A.engine.connect()as H:D[_J]=A._get_row_id_for_key_in_table(L,N[1],H)
			if D[_J]==_A:raise ParentNodeNotFound('Nonexistent parent resource: '+E)
		O=A.schema_path_to_real_table_name[J];P=A.metadata.tables[O]
		with A.engine.begin()as H:Q=H.execute(P.insert().values(**D))
	async def handle_put_config_request(A,data_path,query_dict,request_body,create_callbacks={},change_callbacks={},delete_callbacks={},opaque=_A):
		'\n        Parameters:\n          self: dal\n          data_path: the data_path of the node to change\n          request_body: the new definition (WARNING: modified by call, as lists are stripped out!)\n          callbacks: dicts, may be empty\n          opaque: an object that DAL passes into the callback\n\n          returns boolean for if PUT acted like a POST (i.e., implicit create)\n        ';B=data_path;assert B!='';assert not(B!=_B and B[-1]==_B)
		with A.engine.begin()as D:
			E=await A._handle_put_config_request(B,query_dict,request_body,create_callbacks,change_callbacks,delete_callbacks,opaque,D)
			if A.post_dal_callbacks is not _A:
				for C in A.post_dal_callbacks:
					try:await C[0](C[1],D,C[2])
					except Exception as F:A.post_dal_callbacks=_A;raise F
				A.post_dal_callbacks=_A
			return E
	async def _handle_put_config_request(B,data_path,query_dict,request_body,create_callbacks,change_callbacks,delete_callbacks,opaque,conn):
		"\n        Same as handle_put_config_request() but with the following differences:\n          - no input validation\n          - no starting of a new transaction\n\n        Algorithm:\n          - determine what table it's in\n          - if list, remove row from table (don't need dbpath for this)\n          - else, pop node off jsob and save\n\n          returns boolean for if PUT acted like a POST (i.e., implicit create)\n        ";U=delete_callbacks;S=create_callbacks;P=opaque;L=change_callbacks;I=data_path;F=query_dict;D=request_body;C=conn;K=re.sub(_F,'',I);assert type(D)==dict
		if I==_B:A=B._get_dbpath_for_data_path(_B,ContentType.CONFIG_ANY,C);assert A!=_A;await B.recursive_compare_and_put(A.row_id,_B,D,A.node_ptr,_A,A,S,L,U,P,C);M=B.schema_path_to_real_table_name[A.table_name];J=B.metadata.tables[M];C.execute(J.update().where(J.c.row_id==A.row_id).values(jsob=A.jsob));return _C
		assert len(D)==1;assert K!=_B;A=B._get_dbpath_for_data_path(I,ContentType.CONFIG_ANY,C)
		if A==_A:
			assert I!=_B;G,X=I.rsplit(_B,1)
			if G=='':G=_B
			A=B._get_dbpath_for_data_path(G,ContentType.CONFIG_ANY,C)
			if A==_A:raise ParentNodeNotFound(_b+G+_T)
			await B._handle_post_config_request(A,F,D,S,L,P,C);B._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,C);return _E
		Y=next(iter(D));D=D[Y]
		if type(D)==list:assert len(D)==1;D=D[0]
		await B.recursive_compare_and_put(A.row_id,I,D,A.node_ptr,A.prev_ptr,A,S,L,U,P,C)
		if K in B.config_true_obu_seq_nodes and F!=_A and _O in F:
			if type(A.prev_ptr)==list:A.prev_ptr.remove(A.node_ptr)
			else:assert type(A.prev_ptr)==dict;Z,Q=I.rsplit(_D,1);H=B._get_obu_idx_list_for_list_path(Z,C);H=[A for A in H if A[0][1]!=Q]
			N=F[_O]
			if N==_c:
				assert _K not in F
				if type(A.prev_ptr)==list:A.prev_ptr.insert(0,A.node_ptr)
				else:E=0;O=[(B.table_keys[K],Q),(_H,E)];H.insert(0,O)
			elif N==_U:
				assert _K not in F
				if type(A.prev_ptr)==list:A.prev_ptr.append(A.node_ptr)
				else:E=len(H);O=[(B.table_keys[K],Q),(_H,E)];H.append(O)
			else:
				assert N in(_P,_Q);assert _K in F;V=unquote(F[_K].rsplit(_D,1)[1])
				if type(A.prev_ptr)==list:E=A.prev_ptr.index(V)
				else:
					E=0
					for R in H:
						if R[0][1]==V:break
						E+=1
				if N==_P:0
				elif N==_Q:E=E+1
				if type(A.prev_ptr)==list:A.prev_ptr.insert(E,A.node_ptr)
				else:O=[(B.table_keys[K],Q),(_H,E)];H.insert(E,O)
			if type(A.prev_ptr)!=list:
				for R in range(len(H)):H[R][1]=_H,R
				M=B.schema_path_to_real_table_name[K];J=B.metadata.tables[M];a=B.table_keys[K];b=B._get_row_data_for_list_path(I,C);c=J.update().where(and_(J.c.pid==b[_J],getattr(J.c,a)==bindparam(_R))).values(obu_idx=bindparam(_S));e=C.execute(c,[{_R:A[0][1],_S:A[1][1]}for A in H])
		M=B.schema_path_to_real_table_name[A.table_name];J=B.metadata.tables[M];C.execute(J.update().where(J.c.row_id==A.row_id).values(jsob=A.jsob))
		if K in B.config_true_obu_seq_nodes and F!=_A and _O in F:
			G,X=I.rsplit(_B,1)
			if G=='':G=_B
			W=re.sub(_F,'',G)
			if W in L:
				T=B._get_dbpath_for_data_path(G,ContentType.CONFIG_TRUE,C);assert T!=_A
				for d in L[W]:await d(G,T.jsob,T.jsob_data_path,P)
		return _C
	async def recursive_compare_and_put(C,pid,data_path,req_body_iter,dbpath_curr_ptr,dbpath_prev_ptr,dbpath,create_callbacks,change_callbacks,delete_callbacks,opaque,conn):
		"\n        Compare the incoming request against what is in the database\n          - anything in the request not in the database is added to dbpath_curr_ptr\n          - anything in the database not in the request is removed from dbpath_curr_ptr\n          - terminal values are replace by request's value\n\n        Parameters:\n          - pid: parent table's row-id\n          - data_path: current data_path in tree (MAY not be on a subtable boundary)\n          - req_body_iter: pointer to this place in the path in the request tree\n          - dbpath_curr_ptr: pointer to this place in the path in the database tree\n          - dbpath_prev_ptr: pointer to the previous node in the database tree (only used when 'dbpath_curr_ptr' is a terminal value)\n          - create_callbacks: dict containing 'create' callbacks\n          - change_callbacks: dict containing 'change' callbacks\n          - (NOTE: 'delete_callbacks' not executing inline only because they cannot affect the persisted jsob)\n          - opaque: arbitrary type passed into callbacks\n          - response: (poorly named) dict containing the paths added/changed/deleted\n          - conn: current database connection\n\n        Assumes caller opens dpath and saves remaining jsob as needed.\n\n        Returns a boolean to indicate if the call changed list contents, and hence tells\n        the recursive caller to execute the change callbacks.\n        ";Y=dbpath_prev_ptr;P=pid;N=delete_callbacks;M=change_callbacks;L=create_callbacks;J=opaque;I=conn;H=dbpath;F=dbpath_curr_ptr;D=req_body_iter;B=data_path;assert type(D)==type(F);E=re.sub(_F,'',B)
		if B==_B:0
		if type(D)==dict:
			S=set(list(D.keys()));T=set(list(F.keys()))
			for A in [A for A in S if A not in T]:
				if type(D[A])==list:
					U=_B+A if B==_B else B+_B+A;V=re.sub(_F,'',U)
					if V in C.table_keys:await C.recursive_compare_and_put(P,U,D[A],[],_A,H,L,M,N,J,I);F[A]=[]
					else:F[A]=[];await C.recursive_compare_and_put(P,U,D[A],F[A],F,H,L,M,N,J,I)
				else:
					E=re.sub(_F,'',B);V=_B+A if E==_B else E+_B+A;assert type(D)!=list;G=_B+A if B==_B else B+_B+A;F[A]=D[A];d=await C._recursively_post_subtable_data(P,G,D[A],H.jsob,H.jsob_data_path,L,J,I)
					if d==1:assert type(D)==dict;assert type(D[A])==list;assert len(D[A])==0;D.pop(A);F.pop(A)
			for A in T-S:
				if _C:U=_B+A if B==_B else B+_B+A;await C.recursive_compare_and_put(P,U,[],F[A],F,H,L,M,N,J,I);del F[A]
				else:
					V=_B+A if E==_B else E+_B+A
					if V in C.config_false_prefixes:0
					else:G=_B+A if B==_B else B+_B+A;await C._recursively_delete_subtable_data(H.row_id,G,F[A],N,J,I);del F[A]
			Z=_C
			for A in T&S:
				G=_B+A if B==_B else B+_B+A;V=re.sub(_F,'',G);e=await C.recursive_compare_and_put(P,G,D[A],F[A],F,H,L,M,N,J,I)
				if e==_E:Z=_E
			if T-S or(S-T or Z==_E):
				if E in M:
					for R in M[E]:await R(B,H.jsob,H.jsob_data_path,J)
			return _C
		elif type(D)==list:
			if E in C.table_keys:
				W=[A[C.table_keys[E]]for A in D];O=set(W);K=set([A[0]for A in C._get_keys_in_table_having_pid(E,P,I)])
				for A in [A for A in W if A not in K]:assert B!=_B;G=B+_D+A;f=[B for B in D if B[C.table_keys[E]]==A][0];g=[f];n=await C._recursively_post_subtable_data(P,B,g,_A,_A,L,J,I)
				for A in K-O:G=B+_D+A;o,a=B.rsplit(_B,1);assert a!='';await C._recursively_delete_subtable_data(H.row_id,G,Y[a],N,J,I)
				for A in K&O:G=B+_D+A;Q=C._get_dbpath_for_data_path(G,ContentType.CONFIG_TRUE,I);assert Q!=_A;h=[B for B in D if B[C.table_keys[E]]==A][0];await C.recursive_compare_and_put(Q.row_id,G,h,Q.node_ptr,Q.prev_ptr,Q,L,M,N,J,I);C._update_jsob_for_row_id_in_table(Q.table_name,Q.row_id,Q.jsob,I)
				b=_C
				if E in C.config_true_obu_seq_nodes:
					i=C._get_obu_idx_list_for_list_path(B,I);c=[[(_d,A),(_H,W.index(A))]for A in W]
					if c!=i:j=C.schema_path_to_real_table_name[E];X=C.metadata.tables[j];k=C.table_keys[E];l=X.update().where(and_(X.c.pid==P,getattr(X.c,k)==bindparam(_R))).values(obu_idx=bindparam(_S));p=I.execute(l,[{_R:A[0][1],_S:A[1][1]}for A in c]);b=_E
				if O-K or K-O or b==_E:return _E
				return _C
			else:
				O=set(D);K=set(F);F.clear();F.extend(D)
				for A in O-K:
					G=B+_D+unquote(A)
					if E in L:
						for R in L[E]:await R(G,H.jsob,H.jsob_data_path,J)
				for A in K-O:
					G=B+_D+unquote(A)
					if E in N:
						for R in N[E]:await R(G,J)
				if O-K or K-O:return _E
				return _C
		else:
			if F!=D:
				m=re.sub('^.*/','',B);Y[m]=D
				if E in M:
					for R in M[E]:await R(B,H.jsob,H.jsob_data_path,J)
			else:0
			return _C
		raise NotImplementedError('logic never reaches this point')
	async def handle_put_opstate_request(D,data_path,request_body):
		'\n        Parameters:\n          - the data_path MUST point to a CONFIG FALSE node.\n          - the data_path immediate ancestor (CT or CF) node MUST exist.\n          - the request_body MUST NOT contain any subtable data.\n\n        This routine is expected to handle just one use case: to\n        update/create a single JSOB-bound CONFIG FALSE subtree.\n        \n        It is NOT expected to append a row into an \'opstate\' table (e.g.,\n        audit log, alarms), as the POST-OPSTATE call is used for that.\n\n        The implementations simply inserts or replaces the request_body\n        at the location in the JSOB specified by data_path.\n\n        This routine may be used to create a new/empty parent container\n        for an CONFIG FALSE \'list\' node, which itself would be persisted\n        in another SQL table.\n\n          Comments:\n             - the need to create the parent CONFIG FALSE container\n               would only happen once, as there is no need to ever\n               "update" it thereafter.\n             - perhaps there is a more direct way to create these\n               parent containers?  (DUNNO)\n             - FIXME:\n                 - seems like the change-callback may need to take\n                   a parameter indicating if the change was due to\n                   a POST/PUT/DELETE op\n                 - otherwise these parent containers may be PUT-ed\n                   over and over again\n                     - we only want to create them the FIRST TIME ONLY!!!\n\n        This routine is provided for internal access only.  It is not\n        accessible via the public RESTCONF interface.\n\n        Recursion strategy: None - all nodes must exist in the JSOB.\n        ';F=request_body;B=data_path;assert B!='';assert not(B!=_B and B[-1]==_B)
		with D.engine.begin()as G:
			C,E=B.rsplit(_B,1);assert E!=''
			if C=='':C=_B
			A=D._get_dbpath_for_data_path(C,ContentType.CONFIG_ANY,G)
			if A==_A:raise ParentNodeNotFound(_b+C+_T)
			assert type(A.node_ptr[E])==type(F);A.node_ptr[E]=F;D._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,G)
	async def handle_delete_config_request(A,data_path,delete_callbacks,change_callbacks,opaque):
		'\n        Parameters:\n          self: dal\n          data_path: the data_path of the node to create\n          delete_callbacks: a dict, keyed by schema paths, mapping to a list of callbacks\n          change_callbacks: a dict, keyed by schema paths, mapping to a list of callbacks (only called for parent object)\n        ';B=data_path;assert B!='';assert B!=_B;assert B[-1]!=_B
		with A.engine.begin()as D:
			await A._handle_delete_config_request(B,delete_callbacks,change_callbacks,opaque,D)
			if A.post_dal_callbacks is not _A:
				for C in A.post_dal_callbacks:
					try:await C[0](C[1],D,C[2])
					except Exception as E:A.post_dal_callbacks=_A;raise E
				A.post_dal_callbacks=_A
	async def _handle_delete_config_request(C,data_path,delete_callbacks,change_callbacks,opaque,conn):
		I=opaque;H=change_callbacks;F=conn;D=data_path;E,G=D.rsplit(_B,1)
		if E=='':E=_B
		A=C._get_dbpath_for_data_path(E,ContentType.CONFIG_TRUE,F)
		if A==_A:raise NodeNotFound(_n+D)
		if _D in G:B,O=G.rsplit(_D,1)
		else:B=G
		assert type(A.node_ptr)==dict
		if B not in A.node_ptr:raise NodeNotFound('Cannot delete '+D+'.')
		await C._recursively_delete_subtable_data(A.row_id,D,A.node_ptr[B],delete_callbacks,I,F)
		if type(A.node_ptr[B])==list:
			J=re.sub(_F,'',D)
			if J in C.table_keys:
				L=C._find_rowids_in_table_having_pid(J,A.row_id,F);M=L.fetchall()
				if len(M)==0:assert type(A.node_ptr[B])==list;assert len(A.node_ptr[B])==0;A.node_ptr.pop(B)
			elif len(A.node_ptr[B])==0:A.node_ptr.pop(B)
		else:A.node_ptr.pop(B)
		C._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,F);K=re.sub(_F,'',E)
		if K in H:
			for N in H[K]:await N(E,A.jsob,A.jsob_data_path,I)
	async def _recursively_delete_subtable_data(A,pid,data_path,curr_data_iter,delete_callbacks,opaque,conn):
		'\n        If it weren\'t for executing delete_callbacks, this routine could be (actually "was") much simpler.\n        However, to execute callbacks, we walk the entire data tree, opening sub-jsobs on list boundaries.\n\n        *** delete callbacks are NEVER called on any OPSTATE nodes! ***\n\n        The recursion strategy, on list boundaries, is to iterate descendants before deleting the row \n        (or all rows matching the pid) from a table.\n        \n        Calling routine MUST handle saving out any parent data.\n\n        Parameters:\n          - pid: parent table\'s row-id\n          - data_path: data_path in the request_body tree \n          - curr_data_iter: pointer to this place in the in-database tree\n          - (OLD) prev_data_iter: pointer to \'curr_data_iter\' parent node (for inline callbacks)\n          - delete_callbacks: a dictionary (keyed by schema_path) of callbacks to execute when matching nodes are deleted\n          - opaque: an arbitrary Python object passed into the delete callbacks.delete\n          - conn: the currently opened sqlalchemy database connection\n\n        Returns 1 if a list, 0 otherwise (enables caller to trim dict as needed)\n        ';J=opaque;G=conn;F=delete_callbacks;E=pid;C=data_path;B=curr_data_iter;H=re.sub(_F,'',C)
		if type(B)==list:
			if H in A.table_keys:
				assert B==[];O,K=C.rsplit(_B,1)
				async def L(pid,data_path,delete_callbacks,opaque,conn):
					D=conn;C=data_path;F=re.sub(_F,'',C);B=A._get_dbpath_for_data_path(C,ContentType.CONFIG_TRUE,D)
					if B==_A:raise NodeNotFound(_n+C)
					G=next(iter(B.jsob));H=B.jsob[G];await A._recursively_delete_subtable_data(B.row_id,C,H,delete_callbacks,opaque,D);I=A.schema_path_to_real_table_name[F];E=A.metadata.tables[I];J=D.execute(sa.delete(E).where(E.c.row_id==B.row_id));assert J.rowcount==1
				if _D in K:await L(E,C,F,J,G)
				else:
					P=[B[0]for B in A._get_keys_in_table_having_pid(H,E,G)]
					for D in P:I=C+_D+D;await L(E,I,F,J,G)
			elif any((H.startswith(B)for B in A.config_false_prefixes)):assert B==[];Q=A.schema_path_to_real_table_name[H];M=A.metadata.tables[Q];S=G.execute(sa.delete(M).where(M.c.pid==E))
			else:
				O,K=C.rsplit(_B,1)
				if _D in K:D=unquote(K.rsplit(_D)[1]);assert D in B;I=C;await A._recursively_delete_subtable_data(E,I,D,F,J,G);B.remove(D)
				else:
					while len(B)!=0:
						D=B[0]
						if D is _A:break
						I=C+_D+D;await A._recursively_delete_subtable_data(E,I,D,F,J,G);B.pop(0)
		elif type(B)==dict:
			for N in B.keys():assert C!=_B;I=C+_B+N;await A._recursively_delete_subtable_data(E,I,B[N],F,J,G)
		else:0
		if not type(B)==list and not any((H.startswith(B)for B in A.config_false_prefixes)):
			if H in F:
				for R in F[H]:await R(C,J)
	def _find_rows_in_table_having_pid(E,table_name,pid,query_dict,conn):
		H='limit';G='offset';F='direction';B=query_dict;I=E.schema_path_to_real_table_name[table_name];C=E.metadata.tables[I];A=sa.select(C).where(C.c.pid==pid)
		if _W in B:D=getattr(C.c,B[_W])
		else:D=C.c.row_id
		if F in B and B[F]=='backwards':A=A.order_by(D.desc())
		else:A=A.order_by(D.asc())
		if G in B:A=A.offset(int(B[G]))
		if H in B:A=A.limit(int(B[H]))
		J=conn.execute(A).mappings();return J
	def _find_rowids_in_table_having_pid(B,table_name,pid,conn):C=B.schema_path_to_real_table_name[table_name];A=B.metadata.tables[C];D=conn.execute(sa.select(A.c.row_id).where(A.c.pid==pid).order_by(A.c.row_id));return D
	def _get_keys_in_table_having_pid(A,table_name,pid,conn):B=table_name;D=A.schema_path_to_real_table_name[B];C=A.metadata.tables[D];E=conn.execute(sa.select(getattr(C.c,A.table_keys[B])).where(C.c.pid==pid));return E
	def _get_list_of_direct_subtables_for_schema_path(D,schema_path):
		A=schema_path
		if A!=_B:assert A[-1]!=_B;A+=_B
		C=[]
		for B in sorted(D.schema_path_to_real_table_name.keys()):
			if str(B).startswith(A):
				if not any((A for A in C if str(B).startswith(A+_B))):
					if str(B)!=_B:C.append(str(B))
					else:0
				else:0
		return C
	def _get_row_id_for_key_in_table(B,table_name,key,conn):
		C=table_name;E=B.schema_path_to_real_table_name[C];D=B.metadata.tables[E];F=conn.execute(sa.select(D.c.row_id).where(getattr(D.c,B.table_keys[C])==key));A=F.fetchall();assert A is not _A
		if len(A)==0:return _A
		if len(A)>1:raise TooManyNodesFound()
		return A[0][0]
	def _get_jsob_iter_for_path_in_jsob(D,jsob,path):
		B=path;assert jsob!=_A;assert B[0]!=_B;A=jsob
		if B!='':
			for C in B.split(_B):
				if C!=''and C not in A:return _A
				A=A[C]
				if type(A)==list:assert len(A)==1;A=A[0]
		return A
	def _get_jsob_for_row_id_in_table(C,table_name,row_id,conn):
		'\n          This routine is heavily used by pytests!\n          ...but it\'s also called by _get_dbpath_for_data_path()\n\n          Modified on June 24, 2020 to also return a pseudo "jsob" for pytests\n            - hopefully doesn\'t break other code depending on the assertion for protection\n                - can\'t be the case currently!\n        ';F=row_id;D=table_name;K=C.schema_path_to_real_table_name[D];A=C.metadata.tables[K]
		if D in C.table_keys:E=conn.execute(sa.select(A.c.jsob).where(A.c.row_id==F));G=E.first();assert G!=_A;return G[0]
		else:
			E=conn.execute(sa.select(A).where(A.c.row_id==F)).mappings();H=E.first();assert H!=_A;I=D.rsplit(_B,1)[1];J={I:{}}
			for B in A.c:
				if B.name!=_M and B.name!=_J:J[I][B.name]=H[B.name]
			return J
	def _insert_jsob_into_table(B,pid,table_name,new_jsob,conn,obu_idx=_A):
		'\n          THIS ROUTINE IS ONLY USED BY PYTEST!\n\n          - works for both "config true" and "config false" tables\n        ';F=obu_idx;D=table_name;C=new_jsob;H=B.schema_path_to_real_table_name[D];I=B.metadata.tables[H];E=next(iter(C));A={};A[_J]=pid
		if D in B.table_keys:
			A[B.table_keys[D]]=C[E][B.table_keys[D]]
			if F!=_A:A[_H]=F
			A[_I]=C
		else:
			for G in C[E].keys():A[G]=C[E][G]
		J=conn.execute(I.insert().values(**A));return J.inserted_primary_key[0]
	def _update_jsob_for_row_id_in_table(A,table_name,row_id,new_jsob,conn):C=A.schema_path_to_real_table_name[table_name];B=A.metadata.tables[C];D=conn.execute(sa.update(B).where(B.c.row_id==row_id).values(jsob=new_jsob))
	def _get_table_name_for_schema_path(D,schema_path):
		'\n        Finds closest "config true" or "config false" YANG "list".\n\n        Return the "long" table name that must contain the node identified\n        by "schema_path".  The returned table_name is either shorter or\n        equal to the specified path.  Scans for the longest prefix match.\n        ';B=len(_B);C=_G
		for A in D.schema_path_to_real_table_name.keys():
			if schema_path.startswith(A)and len(A)>B:B=len(A);C=A
		return C
	def _get_obu_idx_list_for_list_path(B,list_path,conn):
		"\n            - a 'list_path' is a data_path that points to a YANG 'list' (not a list entry)\n            - Returns None if not found.\n        ";D=list_path;assert D[0]==_B;assert D!=_B;assert D[-1]!=_B;C='';E=B.global_root_id;K=D[1:].split(_B)
		for F in K:
			if _D in F:
				L,M=F.split(_D);C+=_B+L;I=B._get_table_name_for_schema_path(C);G=B.schema_path_to_real_table_name[I];A=B.metadata.tables[G];J=conn.execute(sa.select(A.c.row_id,A.c.pid).where(and_(A.c.pid==E,getattr(A.c,B.table_keys[I])==M))).mappings();H=J.fetchone()
				if H==_A:return _A
				assert J.fetchone()==_A;assert E==H[_J];E=H[_M]
			else:C+=_B+F
		G=B.schema_path_to_real_table_name[C];A=B.metadata.tables[G];N=conn.execute(sa.select(getattr(A.c,B.table_keys[C]),A.c.obu_idx).where(A.c.pid==E).order_by(A.c.obu_idx));O=N.fetchall();P=[A._mapping.items()._items for A in O];return P
	def _get_row_data_for_list_path(A,data_path,conn):
		"\n            - a 'list_path' is a data_path that points to a YANG 'list' entry.\n            - Returns None if not found.\n        ";B=data_path;assert B[0]==_B;assert B!=_B;assert B[-1]!=_B;G=B[1:].split(_B);assert _D in G[-1];D='';H=A.global_root_id
		for E in G:
			if _D in E:
				K,L=E.split(_D);D+=_B+K;I=A._get_table_name_for_schema_path(D);M=A.schema_path_to_real_table_name[I];C=A.metadata.tables[M];J=conn.execute(sa.select(C.c.row_id,C.c.pid).where(and_(C.c.pid==H,getattr(C.c,A.table_keys[I])==L))).mappings();F=J.fetchone()
				if F==_A:return _A
				assert J.fetchone()==_A;H=F[_M]
			else:D+=_B+E
		return F
	def _get_dbpath_for_data_path(B,data_path,content_type,conn):
		"\n        Returns a DatabasePath object for the data_path, if the item exists\n\n        Calling routine MUST ensure a CONFIG request isn't asking for a CONFIG FALSE node!\n          - test factored out of here for speed\n\n        Calling routine MUST be aware that CONFIG requests will potentially return JSOB with CONFIG FALSE data\n\n        Opstate Considerations:\n          - config false nodes exist in JSOBs too\n          - put the work onto POST operations, as they're relatively rare and don't need to be as performant\n          - all opstate nodes (including non-presence containers) need to be:\n              - created via initial database creation script (e.g. /audit-logs + /alarms)\n              - auto-created via on-change callbacks (e.g., <node>/ref-stats, <device>/lifecycle-stats, <device>/bootstrapping-log, etc.)\n\n        The DatabasePath object:\n          - table_name:  the name of the table\n          - row_id:      the 'id' of the row containing the data\n          - node_ptr:    pointer to the obj addressed by data_path.  for lists, it is the list entry\n          - prev_ptr:    pointer to node_ptr's parent.  used by list-handling logic, as node_ptr is not a python 'list' obj...\n          - inside_path: the path inside the JSOB containing the data\n\n                         ex: '' for the entire object\n                             `GET /` --> ''\n                         ex: 'foo' for a first-level sub-obj\n                             `GET /transport --> 'transport'\n                             `GET /admin-accounts --> 'admin-accounts'\n                         ex: 'foo/bar' for a second-level sub-obj\n                             `GET /transport/listen --> 'transport/listen'\n                             `GET /admin-accounts/admin-account=foo --> 'admin-account' (as it's the top-level of it's jsob)\n\n        ";D=conn;C=data_path;A=DatabasePath();A.data_path=C;A.schema_path=re.sub(_F,'',C);A.table_name=B._get_table_name_for_schema_path(A.schema_path)
		if A.table_name==_A:return _A
		if C!=_B and A.table_name==A.schema_path:
			G,N=C.rsplit(_B,1)
			if _D not in N:
				if G=='':H=_B
				else:H=re.sub(_F,'',G)
				A.table_name=B._get_table_name_for_schema_path(H);assert A.table_name!=_A
		if A.table_name==_G:A.jsob_data_path=_B;A.jsob_schema_path=_B
		else:
			A.jsob_data_path=C;A.jsob_schema_path=re.sub(_F,'',A.jsob_data_path)
			while A.jsob_schema_path!=A.table_name and A.jsob_schema_path!=_B:O=A.jsob_data_path;A.jsob_data_path=re.sub('(.*=[^/]*)/.*','\\g<1>',A.jsob_data_path);assert A.jsob_data_path!=O;A.jsob_schema_path=re.sub(_F,'',A.jsob_data_path)
			assert A.jsob_schema_path!=_B
		if ContentType.CONFIG_FALSE in content_type and any((A.table_name.startswith(C)for C in B.config_false_prefixes)):raise InvalidResourceTarget("RFC 8040 does not allow queries on lists directly and, because SZTPD doesn't support keys on 'config false' lists, it is never possible to query for 'dbpath.table_name' to be returned.  The 'val' layer should've rejected this query... ")
		if A.jsob_schema_path==_B:A.row_id=B.global_root_id
		else:
			assert A.jsob_schema_path in B.table_keys;assert _D in A.jsob_data_path;I=A.jsob_data_path.split(_B);assert _D in I[-1];E=I[-1].split(_D)
			try:A.row_id=B._get_row_id_for_key_in_table(A.table_name,E[1],D)
			except TooManyNodesFound:
				J=B._get_row_data_for_list_path(A.jsob_data_path,D)
				if J is _A:A.row_id=_A
				else:A.row_id=J[_M]
			if A.row_id==_A:return _A
		assert A.data_path.startswith(A.jsob_data_path);K=A.data_path[len(A.jsob_data_path):];assert A.schema_path.startswith(A.jsob_schema_path);L=A.schema_path[len(A.jsob_schema_path):]
		if A.table_name==_G:A.inside_path=A.schema_path[1:]
		else:M=re.findall(_m,A.jsob_data_path);assert len(M)!=0;E=M[-1].split(_D);A.inside_path=E[0];P=re.sub('^'+A.table_name,'',A.schema_path);assert P==L;A.inside_path+=L
		assert A.inside_path==''or A.inside_path[0]!=_B;A.jsob=B._get_jsob_for_row_id_in_table(A.table_name,A.row_id,D);A.node_ptr=A.jsob;A.prev_ptr=_A
		if A.inside_path=='':A.path_segments=[''];return A
		A.path_segments=A.inside_path.split(_B);Q=''
		for F in A.path_segments:
			Q+=_B+F
			if type(A.node_ptr)==list:raise NotImplementedError('is this path used?');A.prev_ptr=A.node_ptr;A.node_ptr=A.node_ptr[0]
			if F not in A.node_ptr:return _A
			else:A.prev_ptr=A.node_ptr;A.node_ptr=A.node_ptr[F]
		if type(A.node_ptr)==list:
			if _D in K:
				R=unquote(K.rsplit(_D,1)[1])
				try:S=A.node_ptr.index(R)
				except:return _A
				A.prev_ptr=A.node_ptr;A.node_ptr=A.node_ptr[S]
		return A
	def _init(A,url,cacert_param,cert_param,key_param):
		K=key_param;J=cert_param;I=cacert_param;G=url
		if not(G.startswith('sqlite:///')or G.startswith(_Z)or G.startswith(_o)):raise SyntaxError('The database url contains an unrecognized dialect.')
		F={}
		if I is not _A:
			F[_N]={};F[_N]['ca']=I
			if J is not _A:F[_N]['cert']=J
			if K is not _A:F[_N][_d]=K
		A.engine=sa.create_engine(G,connect_args=F);A.db_schema=_A;A.table_keys={};A.config_false_prefixes={};A.config_true_obu_seq_nodes={};A.schema_path_to_real_table_name={};A.leafrefs={};A.referers={};A.ref_stat_collectors={}
		try:
			if A.engine.url.database==_e or not db_utils.database_exists(A.engine.url,connect_args=F):A.engine=_A;raise NotImplementedError
		except sa.exc.OperationalError as H:
			if H.orig and'Access denied'in str(H.orig):N=re.sub('^.*"','',re.sub('")$','',str(H.orig)));raise AuthenticationFailed('Authentication failed: '+N)
			else:raise H
		if A.engine.dialect.name==_f:A.schema_path_to_real_table_name[_B]=_G;A.schema_path_to_real_table_name[_G]=_G
		else:
			A.db_schema=A.engine.url.database;A.schema_path_to_real_table_name[_B]=A.db_schema.join(_a);A.schema_path_to_real_table_name[_G]=A.db_schema+_a;O=A.engine.execute(_p);L=O.fetchall();P=[L[A][0]for A in range(len(L))]
			if A.db_schema not in P:A.engine.execute(sa.schema.CreateSchema(A.db_schema));raise NotImplementedError
		A.metadata=sa.MetaData(bind=A.engine,schema=A.db_schema);A.metadata.reflect()
		for Q in A.metadata.tables.values():
			for D in Q.c:
				if type(D.type)is sa.sql.sqltypes.BLOB or type(D.type)is sa.sql.sqltypes.PickleType:D.type=sa.PickleType()
				if A.engine.dialect.name==_Z and type(D.type)is sa.dialects.mysql.types.LONGTEXT:D.type=sa.JSON()
				elif type(D.type)is sa.sql.sqltypes.JSON:D.type=sa.JSON()
		B=A.metadata.tables[A.schema_path_to_real_table_name[_G]]
		with A.engine.connect()as E:
			C=E.execute(sa.select(B.c.jsob).where(B.c.name==_q));M=C.first()[0]
			if M[_g]!=1:raise AssertionError('The database version ('+M[_g]+') is unexpected.')
			C=E.execute(sa.select(B.c.jsob).where(B.c.name==_r));A.app_ns=C.first()[0];C=E.execute(sa.select(B.c.row_id).where(B.c.name==_h));A.global_root_id=C.first()[0];C=E.execute(sa.select(B.c.jsob).where(B.c.name==_s));A.table_keys=C.first()[0];C=E.execute(sa.select(B.c.jsob).where(B.c.name==_t));A.schema_path_to_real_table_name=C.first()[0];C=E.execute(sa.select(B.c.jsob).where(B.c.name==_u));A.config_false_prefixes=C.first()[0];C=E.execute(sa.select(B.c.jsob).where(B.c.name==_v));A.config_true_obu_seq_nodes=C.first()[0]
	def _create(A,url,cacert_param,cert_param,key_param,yl_obj,opaque):
		Q='default startup endpoint';P='endpoint';O='SZTPD_TEST_PATH';N='_sztp_ref_stats_stmt';M='_sztp_globally_unique_stmt';J=key_param;I=cert_param;H=cacert_param;G=yl_obj;C='name';F={}
		if H is not _A:
			F[_N]={};F[_N]['ca']=H
			if I is not _A:F[_N]['cert']=I
			if J is not _A:F[_N][_d]=J
		A.engine=sa.create_engine(url,connect_args=F)
		if A.engine.url.database!=_e and db_utils.database_exists(A.engine.url,connect_args=F):raise AssertionError('Database already exists (call init() first).')
		if A.engine.url.database!=_e:
			if A.engine.dialect.name==_Z:db_utils.create_database(A.engine.url,encoding='utf8mb4',connect_args=F)
			else:db_utils.create_database(A.engine.url,connect_args=F)
		A.db_schema=_A
		if A.engine.dialect.name in(_Z,_o):
			A.db_schema=str(A.engine.url.database);E=A.engine.execute(_p);K=E.fetchall();R=[K[A][0]for A in range(len(K))]
			if A.db_schema not in R:A.engine.execute('CREATE SCHEMA IF NOT EXISTS %s;'%A.db_schema)
		A.metadata=sa.MetaData(schema=A.db_schema);D=sa.Table(_G,A.metadata,sa.Column(_M,sa.Integer,primary_key=_E),sa.Column(_J,sa.Integer,unique=_E),sa.Column(C,sa.String(250),unique=_E),sa.Column(_I,jsob_type));A.metadata.create_all(bind=A.engine)
		with A.engine.begin()as B:B.execute(D.insert(),{C:_q,_I:{_g:1}});B.execute(D.insert(),{C:_r,_I:A.app_ns});B.execute(D.insert(),{C:_k,_I:G});B.execute(D.insert(),{C:'opaque',_I:opaque})
		A.table_keys={_G:C,_B:C};A.config_true_obu_seq_nodes={};A.config_false_prefixes={};A.schema_path_to_real_table_name={};A.leafrefs={};A.referers={};A.ref_stat_collectors={}
		if A.engine.dialect.name==_f:A.schema_path_to_real_table_name[_B]=_G;A.schema_path_to_real_table_name[_G]=_G
		else:A.schema_path_to_real_table_name[_B]=A.db_schema+_a;A.schema_path_to_real_table_name[_G]=A.db_schema+_a
		def S(self,stmt,sctx):self.globally_unique=stmt.argument
		setattr(yangson.schemanode.SchemaNode,M,S);yangson.schemanode.SchemaNode._stmt_callback['wn-app:globally-unique']=M
		def T(self,stmt,sctx):self.ref_stats=stmt.argument
		setattr(yangson.schemanode.SchemaNode,N,T);yangson.schemanode.SchemaNode._stmt_callback['wn-app:ref-stats']=N;U=re.sub('\\..*','',__name__);L=pkg_resources.resource_filename(U,'yang/')
		if os.environ.get(O):A.dm=yangson.DataModel(json.dumps(G),[os.environ.get(O),L])
		else:A.dm=yangson.DataModel(json.dumps(G),[L])
		A._gen_tables(A.dm.schema,_G)
		with A.engine.begin()as B:E=B.execute(D.insert(),{C:_t,_I:A.schema_path_to_real_table_name});E=B.execute(D.insert(),{C:_s,_I:A.table_keys});E=B.execute(D.insert(),{C:_u,_I:A.config_false_prefixes});E=B.execute(D.insert(),{C:_v,_I:A.config_true_obu_seq_nodes})
		if os.environ.get('SZTPD_TEST_DAL'):
			with A.engine.begin()as B:E=B.execute(D.insert().values(name=_h,jsob={}))
			A.global_root_id=E.inserted_primary_key[0]
		else:
			with A.engine.begin()as B:E=B.execute(D.insert().values(name=_h,jsob={A.app_ns+':transport':{'listen':{P:[]}},A.app_ns+':audit-log':{'log-entry':[]}}))
			A.global_root_id=E.inserted_primary_key[0];V=A.schema_path_to_real_table_name[_B+A.app_ns+':transport/listen/endpoint'];W=A.metadata.tables[V]
			with A.engine.begin()as B:X=B.execute(W.insert().values(pid=A.global_root_id,name=Q,jsob={P:{C:Q,'use-for':'wn-app:native-interface','http':{'tcp-server-parameters':{'local-address':os.environ.get('SZTPD_INIT_ADDR','127.0.0.1')}}}}))
	def _gen_tables(D,node,parent_table_name):
		"\n        Creates a SQL table for *every* YANG 'list' node.\n          - analysis shows that this leads to about a 25% increase over the number of hand-coded tables,\n            but is simpler than using lots of custom annotations...\n\n        ";N='ref_stats';G=parent_table_name;B=node
		if issubclass(type(B),yangson.schemanode.ListNode):
			C=[];C.append(sa.Column(_M,sa.Integer,primary_key=_E));O=D.schema_path_to_real_table_name[G];C.append(sa.Column(_J,sa.Integer,sa.ForeignKey(O+'.row_id'),index=_E,nullable=_C))
			if B.config==_E:
				if len(B.keys)>1:raise NotImplementedError('Only supports lists with at most one key.')
				E=B.get_child(*B.keys[0]);D.table_keys[B.data_path()]=E.name
				if type(E.type)==yangson.datatype.StringType:C.append(sa.Column(E.name,sa.String(250),nullable=_C))
				elif type(E.type)==yangson.datatype.Uint32Type:C.append(sa.Column(E.name,sa.Integer,nullable=_C))
				elif type(E.type)==yangson.datatype.IdentityrefType:C.append(sa.Column(E.name,sa.String(250),nullable=_C))
				elif type(E.type)==yangson.datatype.UnionType:C.append(sa.Column(E.name,sa.String(250),nullable=_C))
				elif type(E.type)==yangson.datatype.LeafrefType:C.append(sa.Column(E.name,sa.String(250),nullable=_C))
				else:raise Exception('Unsupported key node type: '+str(type(E.type)))
				if hasattr(E,'globally_unique'):C.append(sa.UniqueConstraint(E.name))
				else:C.append(sa.UniqueConstraint(E.name,_J))
				if B.user_ordered==_E:C.append(sa.Column(_H,sa.Integer,index=_E,nullable=_C))
				C.append(sa.Column(_I,jsob_type,nullable=_C))
			else:
				assert B.config==_C;assert hasattr(B,N)==_C
				for A in B.children:
					if issubclass(type(A),yangson.schemanode.LeafNode):
						if type(A.type)==yangson.datatype.StringType:
							if str(A.type)=='date-and-time(string)':C.append(sa.Column(A.name,sa.DateTime,index=_E,nullable=A.mandatory==_C or A.when!=_A))
							else:C.append(sa.Column(A.name,sa.String(250),index=_E,nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.Uint16Type:C.append(sa.Column(A.name,sa.SmallInteger,index=_E,nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.InstanceIdentifierType:C.append(sa.Column(A.name,sa.String(250),nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.LeafrefType:C.append(sa.Column(A.name,sa.String(250),nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.IdentityrefType:C.append(sa.Column(A.name,sa.String(250),nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.EnumerationType:C.append(sa.Column(A.name,sa.String(250),index=_E,nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.UnionType:
							J=0;S=_E
							for K in A.type.types:
								if issubclass(type(K),yangson.datatype.StringType):J+=1
								else:raise Exception('Unhandled union type: '+str(type(K)))
							if J==len(A.type.types):C.append(sa.Column(A.name,sa.String(250),index=_E,nullable=A.mandatory==_C or A.when!=_A))
							else:raise Exception('FIXME: not all union subtypes are stringafiable')
						else:raise Exception('Unhandled leaf type: '+str(type(A.type)))
					elif issubclass(type(A),yangson.schemanode.ChoiceNode):
						H=_E
						for I in A.children:
							assert type(I)==yangson.schemanode.CaseNode
							if len(I.children)>1:H=_C;break
							else:
								for P in I.children:
									if type(P)!=yangson.schemanode.LeafNode:H=_C;break
						if H==_E:C.append(sa.Column(A.name,sa.String(250),index=_E,nullable=A.mandatory==_C or A.when!=_A))
						else:C.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.AnydataNode):C.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.LeafListNode):C.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.ListNode):C.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.ContainerNode):C.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.NotificationNode):0
					else:raise Exception('Unhandled list child type: '+str(type(A)))
			Q=re.sub('^/.*:','',B.data_path()).split(_B)
			if D.engine.dialect.name==_f:F=''
			else:F=D.db_schema+'.'
			for R in Q:F+=R[0]
			while F in D.schema_path_to_real_table_name.values():F+='2'
			D.schema_path_to_real_table_name[B.data_path()]=F
			if D.db_schema is _A:L=sa.Table(F,D.metadata,*(C))
			else:L=sa.Table(re.sub('^'+D.db_schema+'.','',F),D.metadata,*(C))
			L.create(bind=D.engine);G=B.data_path()
		if B.config==_C and issubclass(type(B),yangson.schemanode.DataNode):
			M=B.data_path()
			if not any((M.startswith(A)for A in D.config_false_prefixes)):D.config_false_prefixes[M]=_E
		if B.config==_E and issubclass(type(B),yangson.schemanode.SequenceNode)and B.user_ordered==_E:D.config_true_obu_seq_nodes[B.data_path()]=_A
		if hasattr(B,N):D.ref_stat_collectors[B.data_path()]=_A
		if issubclass(type(B),yangson.schemanode.InternalNode):
			if not(type(B)==yangson.schemanode.ListNode and B.config==_C)and not type(B)==yangson.schemanode.RpcActionNode and not type(B)==yangson.schemanode.NotificationNode:
				for A in B.children:D._gen_tables(A,G)



# Copyright (c) 2023 Watsen Networks.  All Rights Reserved.

_S='content'
_R='certificates'
_Q='certificate'
_P='import'
_O='\\g<1>'
_N='YYYY-MM-DD'
_M='module'
_L='ietf-yang-library:modules-state'
_K='contentType'
_J='encoding = '
_I='utf-8'
_H='implement'
_G='xml'
_F='json'
_E='conformance-type'
_D='namespace'
_C='name'
_B='revision'
_A=None
import re,sys,pem,json,base64,yangson,datetime,textwrap,traceback,pkg_resources,xml.etree.ElementTree as ET
from os import listdir
from enum import Enum
from aiohttp import web
from xml.dom import minidom
from pyasn1.type import tag
from urllib.parse import unquote
from pyasn1_modules import rfc5652
from pyasn1_modules import rfc5280
from pyasn1.codec.der import decoder as der_decoder
from pyasn1.codec.der import encoder as der_encoder
from yangson.xmlparser import XMLParser
class RedundantQueryParameters(Exception):0
class MalformedDataPath(Exception):0
app_name=re.sub('\\..*','',__name__)
def set_yyyy_mm_dd(yl_obj):
	"\n      Replace 'YYYY-MM-DD' revision strings with correct date given\n      YANG filenames.  All this because 'work-in-progress' YANG\n      modules tend to change daily.\n    ";B=yl_obj;D=pkg_resources.resource_filename(app_name,'yang')
	for A in B[_L][_M]:
		if A[_B]==_N:
			for C in listdir(D):
				if C.startswith(A[_C]):E=re.sub(A[_C]+'@(.*)\\.yang',_O,C);A[_B]=E;break
	return B
yl4errors={_L:{'module-set-id':'TBD',_M:[{_C:'ietf-yang-types',_B:'2013-07-15',_D:'urn:ietf:params:xml:ns:yang:ietf-yang-types',_E:_P},{_C:'ietf-restconf',_B:'2017-01-26',_D:'urn:ietf:params:xml:ns:yang:ietf-restconf',_E:_H},{_C:'ietf-netconf-acm',_B:'2018-02-14',_D:'urn:ietf:params:xml:ns:yang:ietf-netconf-acm',_E:_P},{_C:'ietf-yang-structure-ext',_B:'2020-06-22',_D:'urn:ietf:params:xml:ns:yang:ietf-yang-structure-ext',_E:_H},{_C:'ietf-crypto-types',_B:_N,_D:'urn:ietf:params:xml:ns:yang:ietf-crypto-types',_E:_H}]}}
path=pkg_resources.resource_filename(app_name,'yang')
path4errors=pkg_resources.resource_filename(app_name,'yang4errors')
yl4errors=set_yyyy_mm_dd(yl4errors)
dm4errors=yangson.DataModel(json.dumps(yl4errors),[path4errors,path])
def gen_rc_errors(error_type,error_tag,error_app_tag=_A,error_path=_A,error_message=_A,error_info=_A):
	'\n       Returns the ietf-restconf:errors object.\n\n       Note: Was going to return a web.Response object, but setting the HTTP status code\n             by "error-tag" is context specific (e.g.: is "access-denied" 401 or 403).\n    ';E=error_info;D=error_message;C=error_path;B=error_app_tag;A={};A['error-type']=error_type;A['error-tag']=error_tag
	if B is not _A:A['error-app-tag']=B
	if C is not _A:A['error-path']=C
	if D is not _A:A['error-message']=D
	if E is not _A:A['error-info']=E
	return{'ietf-restconf:errors':{'error':[A]}}
def enc_rc_errors(encoding,errors_obj):
	'\n       Returns the ietf-restconf:errors in the specied encoding.\n    ';B=errors_obj;A=encoding;print('enc_rc_errors: '+A)
	if A==_F:return json.dumps(B,indent=2)
	if A==_G:C=dm4errors.from_raw(B);D=C.to_xml();E=ET.tostring(D).decode(_I);F=minidom.parseString(E);return F.toprettyxml(indent='  ')
	raise NotImplementedError(_J+A)
class Encoding(Enum):json=1;xml=2
def obj_to_encoded_str(obj,enc,dm,sn,strip_wrapper=False):
	'\n       Serialize object to string using specified encoding (no validation)\n\n       Parameters:\n         obj: the Python object to be encoded\n         enc: the encoding to use\n         dm:  the Yangson DataModel object to use\n         sn:  the Yangson SchemaNode object to use\n         strip_wrapper: for XML only, indicates if the Yangson added <content-data> wrapper should be removed\n\n       Returns:\n         a string for the encoded object\n\n       Exceptions:\n         TBD\n    ';B=enc
	if B==Encoding.json:return json.dumps(obj,indent=2)
	if B==Encoding.xml:
		C=sn.from_raw(obj);D=yangson.instance.RootNode(C,sn,dm.schema_data,C.timestamp);A=D.to_xml()
		if strip_wrapper==True:assert len(A)==1;A=A[0]
		E=ET.tostring(A).decode(_I);F=minidom.parseString(E);return F.toprettyxml(indent='  ')
	raise NotImplementedError(_J+B)
def encoded_str_to_obj(estr,enc,dm,sn):
	'\n       Deserialize string to object...and validate it!\n\n       Parameters:\n         estr: the encoded string to deserialized\n         enc:  the encoding to use\n         sn:   the Yangson DataModel object to use\n         ch:   the name of a descendent node in both\n               the schema node and the deserialized string\n\n       Returns:\n         an obj for the deserialized string\n\n       Exceptions:\n         TBD\n    ';C="Doesn't match schema: "
	if enc==Encoding.json:
		try:D=json.loads(estr)
		except Exception as A:raise Exception('JSON malformed: '+str(A))
		try:B=sn.from_raw(D)
		except Exception as A:raise Exception(C+str(A))
	elif enc==Encoding.xml:
		try:E=XMLParser(estr);F=E.root
		except Exception as A:raise Exception('XML malformed: '+str(A))
		try:B=sn.from_xml(F)
		except Exception as A:raise Exception(C+str(A))
	else:raise NotImplementedError(_J+encoding)
	try:G=yangson.instance.RootNode(B,sn,dm.schema_data,B.timestamp)
	except Exception as A:raise Exception(C+str(A))
	try:H=G.raw_value()
	except Exception as A:raise Exception('Error transcoding: '+str(A))
	return H
def multipart_pem_to_der_dict(multipart_pem):
	'\n      Convert a dictionary containing DERs of different types\n      to a multipart PEM string.\n\n      Parameters:\n        multipart_pem - a \'str\' containing one or more PEM blocks.\n\n      Returns a \'dict\' keyed by the content type identifiers found\n      in the multipart PEM.\n\n      Content identifiers are those strings found after the "BEGIN"\n      and "END" markers (e.g., CERTIFICATE).\n    ';A={};E=pem.parse(bytes(multipart_pem,_I))
	for F in E:
		C=F.as_text().splitlines();D=base64.b64decode(''.join(C[1:-1]));B=re.sub('-----BEGIN (.*)-----',_O,C[0])
		if B not in A:A[B]=[D]
		else:A[B].append(D)
	return A
def der_dict_to_multipart_pem(der_dict):
	'\n      Convert a dictionary containing DERs of different types\n      to a multipart PEM string.\n\n      Parameters:\n        der_dict - a \'dict\' keyed by content type identifiers.\n\n      Returns a \'str\' object that having a terminating \'\n\'\n      character (Note that the sslContext parameters use \'str\',\n      NOT \'bytes\')\n\n      Content identifiers are those strings commonly found after\n      the "BEGIN" and "END" markers (e.g., CERTIFICATE).\n\n      Currently, only "CERTIFICATE" is supported.\n    ';D='-----\n';C=der_dict;A='';E=C.keys()
	for B in E:
		F=C[B]
		for G in F:H=base64.b64encode(G).decode('ASCII');A+='-----BEGIN '+B+D;A+=textwrap.fill(H,64)+'\n';A+='-----END '+B+D
	return A
def ders_to_degenerate_cms_obj(cert_ders):
	B=rfc5652.CertificateSet().subtype(implicitTag=tag.Tag(tag.tagClassContext,tag.tagFormatSimple,0))
	for E in cert_ders:F,G=der_decoder.decode(E,asn1Spec=rfc5280.Certificate());assert not G;D=rfc5652.CertificateChoices();D[_Q]=F;B[len(B)]=D
	A=rfc5652.SignedData();A['version']=1;A['digestAlgorithms']=rfc5652.DigestAlgorithmIdentifiers().clear();A['encapContentInfo']['eContentType']=rfc5652.id_data;A[_R]=B;C=rfc5652.ContentInfo();C[_K]=rfc5652.id_signedData;C[_S]=der_encoder.encode(A);return C
def degenerate_cms_obj_to_ders(cms_obj):
	A=cms_obj
	if A[_K]!=rfc5652.id_signedData:raise KeyError('unexpected content type: '+str(A[_K]))
	D,H=der_decoder.decode(A[_S],asn1Spec=rfc5652.SignedData());E=D[_R];B=[]
	for F in E:C=F[_Q];assert type(C)==rfc5280.Certificate;G=der_encoder.encode(C);B.append(G)
	return B
def parse_raw_path(full_raw_path):
	'\n       This routine is used by the various "handle_*_request" routines.\n\n       According to https://tools.ietf.org/html/rfc8040#section-3.5.3, any\n       reserved characters in a resource identifier MUST be percent-encoded.\n\n       Resource identifiers MAY appear in URLs as list-keys and as the value\n       for the "point" query parameter.\n\n       Thusly, the client\'s app-level logic SHOULD percent-encode such values,\n       and then the client\'s infra-level logic SHOULD percent-encode remaining\n       values, e.g., UTF-8 could be used for node names.\n\n       Standard URL decoders do not understand this and thus can have parsing\n       errors and/or decode more than they should.\n\n       This routine parses the raw URL path (including query params) and\n       returns the correct data-path and query-params (e.g., the client\'s\n       app-level values).\n    ';N="' appears more than once.  RFC 8040, Section 4.8 states that each parameter can appear at most once.";M="Query parameter '";L='?';G=full_raw_path;D='=';A='/'
	if L in G:assert G.count(L)==1;B,J=G.split(L)
	else:B=G;J=_A
	if B=='':B=A
	elif B[0]!=A:raise MalformedDataPath("The datastore-specific part of the path, when present, must begin with a '/' character.")
	elif B[-1]==A:raise MalformedDataPath("Trailing '/' characters are not supported.")
	if B==A:H=A
	else:
		H='';O=B[1:].split(A)
		for E in O:
			if E=='':raise MalformedDataPath("The data path contains a superflous '/' character.")
			if D in E:assert E.count(D)==1;C,K=E.split(D);H+=A+unquote(C)+D+K
			else:H+=A+unquote(E)
	F=dict()
	if J is not _A:
		P=J.split('&')
		for I in P:
			if D in I:
				C,K=I.split(D,1)
				if C in F:raise RedundantQueryParameters(M+C+N)
				F[unquote(C)]=K
			else:
				if I in F:raise RedundantQueryParameters(M+C+N)
				F[unquote(I)]=_A
	return H,F
def check_http_headers(request,supported_media_types,accept_required):
	'\n      parameters:\n        - request: the incoming web request object\n        - supported_media_types: a tuple containing the media-types supported\n            - e.g., (\'application/yang-data+json\', \'application/yang-data+xml\')\n        - accept_required: boolean for if an \'Accept\' parameter must\'ve been passed (only False when a 204 is expected)\n\n      returns:\n        on error: (web-response object, error_msg *OR* error_obj)\n        on success: (string containing response_media_type, None)\n                      - note that the response_media_type could be "text/plain"\n    ';Q='application/yang-data+xml';P='The request method (';O=accept_required;N='missing-attribute';M='".';L='" or "';K='protocol';I='application/yang-data+json';H='Content-Type';G='text/plain';F=supported_media_types;D='Accept';B=request;assert type(F)==tuple
	if not B.body_exists:
		if any((B.method==A for A in('PUT','PATCH')))or B.method=='POST'and'ietf-datastores:running'in B.path:
			A=web.Response(status=400);C=P+B.method+') must include a request body.'
			if not D in B.headers or not any((B.headers[D]==A for A in F)):A.content_type=G;A.text=C;return A,C
			else:
				A.content_type=B.headers[D];E=gen_rc_errors(K,N,error_message=C)
				if A.content_type==I:A.text=enc_rc_errors(_F,E)
				else:A.text=enc_rc_errors(_G,E)
				return A,E
	if B.body_exists:
		if any((B.method==A for A in('GET','DELETE'))):
			A=web.Response(status=400);C=P+B.method+') should not include a request body. (enforced here)'
			if not D in B.headers or not any((B.headers[D]==A for A in F)):A.content_type=G;A.text=C;return A,C
			else:
				A.content_type=B.headers[D];E=gen_rc_errors(K,N,error_message=C)
				if A.content_type==I:A.text=enc_rc_errors(_F,E)
				else:A.text=enc_rc_errors(_G,E)
				return A,E
		if not H in B.headers:
			A=web.Response(status=400);C='A "Content-Type" value must be specified when a request body is passed. The "Content-Type" value must be "application/yang-data+json" or "application/yang-data+xml".'
			if not D in B.headers or not any((B.headers[D]==A for A in F)):A.content_type=G;A.text=C;return A,C
			else:
				A.content_type=B.headers[D];E=gen_rc_errors(K,N,error_message=C)
				if A.content_type==I:A.text=enc_rc_errors(_F,E)
				else:A.text=enc_rc_errors(_G,E)
				return A,E
		if not any((B.headers[H]==A for A in F)):
			A=web.Response(status=400);C='The "Content-Type" value, when specified, must be "'+L.join(F)+'". Got: "'+B.headers[H]+M
			if not D in B.headers or not any((B.headers[D]==A for A in(I,Q))):A.content_type=G;A.text=C;return A,C
			else:
				A.content_type=B.headers[D];E=gen_rc_errors(K,'bad-attribute',error_message=C)
				if A.content_type==I:A.text=enc_rc_errors(_F,E)
				else:A.text=enc_rc_errors(_G,E)
				return A,E
	J=_A
	if D not in B.headers or B.headers[D]=='*/*':
		if H not in B.headers:
			if O:A=web.Response(status=406);C='Unable to determine response encoding; neither "Accept" nor "Content-Type" specified.  An "Accept" value should be specified, and have the value "'+L.join(F)+M;A.content_type=G;A.text=C;return A,C
			J=G
		elif not any((B.headers[H]==A for A in(I,Q))):
			if O:A=web.Response(status=406);C='Unable to determine response encoding; "Accept" not specified and the "Content-Type" specified ('+B.headers[H]+') is invalid.  An "Accept" value should be specified, and have the value "'+L.join(F)+M;A.content_type=G;A.text=C;return A,C
			J=G
		else:J=B.headers[H]
	elif not any((B.headers[D]==A for A in F)):A=web.Response(status=406);C='Unable to determine response encoding; the "Accept" value specified ('+B.headers[D]+') is invalid.  The "Accept" value, when specified, must have the value "'+L.join(F)+M;A.content_type=G;A.text=C;return A,C
	else:J=B.headers[D]
	return J,_A



# Copyright (c) 2023 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
_B='/ds/ietf-datastores:operational'
_A='/ds/ietf-datastores:running'
import os,ssl,json,base64,pyasn1,yangson,datetime,tempfile,basicauth
from .  import utils
from aiohttp import web
from .handler import RouteHandler
from pyasn1.type import univ
from pyasn1_modules import rfc3447
from pyasn1_modules import rfc5280
from pyasn1_modules import rfc5652
from pyasn1_modules import rfc5915
from pyasn1.codec.der.decoder import decode as der_decoder
from pyasn1.codec.der.encoder import encode as der_encoder
'\n   DESIGN\n\n   for read-only config/opstate requests:\n     - first check if user is allowed to read the node\n     - get RW-lock(READ)\n     - get operational.RW-lock(WRITE)\n     - insert a record into audit log\n     - release operational.RW-lock(WRITE)\n     - fetch node from DAL\n     - release RW-lock(READ)\n     - prune sub-nodes from result as needed\n     - return result\n\n   for read-write config requests:\n     - first check if user is allowed to write the node\n     - get running.RW-lock(WRITE)\n     - get operational.RW-lock(WRITE)\n     - insert a record into audit log\n     - release operational.RW-lock(WRITE)\n     - validate the request via VAL\n     - execute pre-persist callbacks (e.g., crypt-hash)\n     - persist the request via DAL (into <running>)\n     - spawn sub-task (see below) - should block on trying to acquire read lock\n     - release running.RW-lock(WRITE)\n     - return\n     - as a spawned subtask\n       - get running.RW-lock(READ)  // warning, there\'s a lock-gap here!\n       - execute post-persist callbacks (e.g., HUP)\n           - NOTE: "HUP" callback SHOULD be built-in!\n           - FIXME: traverse leafrefs to notify all stakeholders\n                    (e.g., addr-book obj has cascading impact)\n       - FUTURE\n          - get running.RW-lock(READ)\n          - get intended.RW-lock(WRITE)\n          - perform impact-analysis (i.e., traverse leafrefs)\n             - do minimal <running> --> <intended> transformations.\n          - persist <intended>\n          - release intended.RW-lock(WRITE)\n          - release running.RW-lock(READ)\n       - release running.RW-lock(READ)\n\n    for rpcs/actions:\n     - first check if user is allowed to write the node\n     - get RW-lock(READ)\n     - get operational.RW-lock(WRITE)\n     - insert a record into audit log\n     - release operational.RW-lock(WRITE)\n     - validate the request via VAL\n     - execute registed rpc/action callback\n        - FIXME: how can they affect <operational> !!!\n     - release RW-lock(READ)\n     - return result\n\n\n\n   FIXME: <OPERATIONAL>\n\n     - how will <operational> ever be implemented?\n     - ignore for now (just return <running>)\n\n\n\n   FIXME: OPTIMIZED SUB-TENANT ACCESS\n\n     The default RESTCONF server instance represents native yang-library (e.g., 1 vs x)\n  \n     When running multi-tenant, the default would be for sub-tenants to have to access via\n     the "/tenants/tenant[name=\'\']/..." hierarchy.  This might be okay for some, and perhaps\n     necessary for "admins" that have access to more than one sub-tenant (but are not host-\n     system admins), but a more PROFESSIONAL scenario would be to have a PROXY RC server\n     running on another PORT that presents the "single-tenant" view (inc. yang-library)\n     while actually transparently switching to the corrent sub-tenant hierarchy based on\n     which tenant the admin account is mapped to.\n  \n     This is preferable because 1) it leaks less information to sub-tenants (they don\'t\n     know that they\'re sub-tenants) and 2) the admin port can configured to only be\n     accessible from the inside network (security).\n  \n     For now, just the plain-access is being implemented, but the proxy-access should\n     be implemented later!\n\n\n'
async def set_server_header(request,response):response.headers['Server']='<redacted>'
class RestconfServer:
	root='/restconf';prefix_running=root+_A;prefix_operational=root+_B;prefix_operations=root+'/operations';len_prefix_running=len(prefix_running);len_prefix_operational=len(prefix_operational);len_prefix_operations=len(prefix_operations)
	def __init__(A,loop,dal,endpoint_config,view_handler):
		x='client-certs';w='local-truststore-reference';v='ca-certs';u='client-authentication';t='\n-----END CERTIFICATE-----\n';s='-----BEGIN CERTIFICATE-----\n';r='cert-data';q='private-key-format';p=':keystore/asymmetric-keys/asymmetric-key=';o='reference';n='server-identity';m='local-port';l='http';W='ASCII';V=':asymmetric-key';U=None;T='tcp-server-parameters';M='certificate';L='tls-server-parameters';K='/ds/ietf-datastores:running{tail:.*}';J='/';G=dal;D='https';C=endpoint_config;B=view_handler;A.len_prefix_running=len(A.root+_A);A.len_prefix_operational=len(A.root+_B);A.loop=loop;A.dal=G;A.name=C['name'];A.view_handler=B;A.app=web.Application(client_max_size=1024*1024*32);A.app.on_response_prepare.append(set_server_header);A.app.router.add_get('/.well-known/host-meta',A.handle_get_host_meta);A.app.router.add_get(A.root,B.handle_get_restconf_root);A.app.router.add_get(A.root+J,B.handle_get_restconf_root);A.app.router.add_get(A.root+'/yang-library-version',B.handle_get_yang_library_version);A.app.router.add_get(A.root+'/ds/ietf-datastores:operational{tail:.*}',B.handle_get_opstate_request);A.app.router.add_get(A.root+K,B.handle_get_config_request);A.app.router.add_put(A.root+K,B.handle_put_config_request);A.app.router.add_post(A.root+K,B.handle_post_config_request);A.app.router.add_delete(A.root+K,B.handle_delete_config_request);A.app.router.add_post(A.root+'/ds/ietf-datastores:operational/{tail:.*}',B.handle_action_request);A.app.router.add_post(A.root+'/operations/{tail:.*}',B.handle_rpc_request)
		if l in C:F=l
		else:assert D in C;F=D
		A.local_address=C[F][T]['local-address'];A.local_port=os.environ.get('SZTPD_INIT_PORT',8080)
		if m in C[F][T]:A.local_port=C[F][T][m]
		E=U
		if F==D:
			X=C[D][L][n][M][o]['asymmetric-key'];N=A.dal.handle_get_config_request(J+A.dal.app_ns+p+X);O=A.loop.run_until_complete(N);P=O[A.dal.app_ns+V][0]['cleartext-private-key'];Y=base64.b64decode(P)
			if O[A.dal.app_ns+V][0][q]=='ietf-crypto-types:ec-private-key-format':Q,y=der_decoder(Y,asn1Spec=rfc5915.ECPrivateKey());z=der_encoder(Q);Z=base64.b64encode(z).decode(W);assert P==Z;a='-----BEGIN EC PRIVATE KEY-----\n'+Z+'\n-----END EC PRIVATE KEY-----\n'
			elif O[A.dal.app_ns+V][0][q]=='ietf-crypto-types:rsa-private-key-format':Q,y=der_decoder(Y,asn1Spec=rfc3447.RSAPrivateKey());A0=der_encoder(Q);b=base64.b64encode(A0).decode(W);assert P==b;a='-----BEGIN RSA PRIVATE KEY-----\n'+b+'\n-----END RSA PRIVATE KEY-----\n'
			else:raise NotImplementedError('this line can never be reached')
			A1=C[D][L][n][M][o][M];N=A.dal.handle_get_config_request(J+A.dal.app_ns+p+X+'/certificates/certificate='+A1);A2=A.loop.run_until_complete(N);A3=A2[A.dal.app_ns+':certificate'][0][r];A4=base64.b64decode(A3);A5,c=der_decoder(A4,asn1Spec=rfc5652.ContentInfo());A6=A5.getComponentByName('content');A7,c=der_decoder(A6,asn1Spec=rfc5652.SignedData());d=A7.getComponentByName('certificates');H=''
			for A8 in range(len(d)):
				e=d[A8][0]
				for f in e['tbsCertificate']['extensions']:
					if f['extnID']==rfc5280.id_ce_basicConstraints:A9,c=der_decoder(f['extnValue'],asn1Spec=rfc5280.BasicConstraints())
				AA=der_encoder(e);g=base64.b64encode(AA).decode(W)
				if A9['cA']==False:H=s+g+t+H
				else:H+=s+g+t
			E=ssl.create_default_context(ssl.Purpose.CLIENT_AUTH);E.verify_mode=ssl.CERT_OPTIONAL
			with tempfile.TemporaryDirectory()as h:
				i=h+'key.pem';j=h+'certs.pem'
				with open(i,'w')as AB:AB.write(a)
				with open(j,'w')as AC:AC.write(H)
				E.load_cert_chain(j,i)
			if u in C[D][L]:
				I=C[D][L][u]
				def k(truststore_ref):
					C=G.handle_get_config_request(J+G.app_ns+':truststore/certificate-bags/certificate-bag='+truststore_ref);D=A.loop.run_until_complete(C);B=[]
					for E in D[G.app_ns+':certificate-bag'][0][M]:F=base64.b64decode(E[r]);H,I=der_decoder(F,asn1Spec=rfc5652.ContentInfo());assert not I;B+=utils.degenerate_cms_obj_to_ders(H)
					return B
				R=[]
				if v in I:S=I[v][w];R+=k(S)
				if x in I:S=I[x][w];R+=k(S)
				AD=utils.der_dict_to_multipart_pem({'CERTIFICATE':R});E.load_verify_locations(cadata=AD)
		if F==D:assert not E is U
		else:assert E is U
		A.runner=web.AppRunner(A.app);A.loop.run_until_complete(A.runner.setup());A.site=web.TCPSite(A.runner,host=A.local_address,port=A.local_port,ssl_context=E,reuse_port=True);A.loop.run_until_complete(A.site.start())
	async def handle_get_host_meta(B,request):'\n          no auth check, since outside restconf server.\n          no audit log entry, since outside restconf server.\n        ';A=web.Response();A.content_type='application/xrd+xml';A.text='<XRD xmlns="http://docs.oasis-open.org/ns/xri/xrd-1.0">\n  <Link rel="restconf" href="/restconf"/>\n</XRD>';return A



# Copyright (c) 2023 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
_Ab='Returning an RPC-error provided by callback (NOTE: RPC-error != exception, hence a normal exit).'
_Aa='Unrecognized error-tag: '
_AZ='partial-operation'
_AY='operation-failed'
_AX='rollback-failed'
_AW='data-exists'
_AV='resource-denied'
_AU='lock-denied'
_AT='unknown-namespace'
_AS='bad-element'
_AR='unknown-attribute'
_AQ='bad-attribute'
_AP='missing-attribute'
_AO='exception-thrown'
_AN='functions'
_AM='callback-details'
_AL='from-device'
_AK='identity-certificate'
_AJ='source-ip-address'
_AI='serial-number'
_AH='"ietf-sztp-bootstrap-server:input" is missing.'
_AG='/ietf-sztp-bootstrap-server:report-progress'
_AF='Resource does not exist.'
_AE='Requested resource does not exist.'
_AD=':log-entry'
_AC='/devices/device='
_AB=':devices/device='
_AA='2019-04-30'
_A9='urn:ietf:params:xml:ns:yang:ietf-yang-types'
_A8='ietf-yang-types'
_A7='module-set-id'
_A6='ietf-yang-library:modules-state'
_A5='application/yang-data+xml'
_A4='webhooks'
_A3='callout-type'
_A2='passed-input'
_A1='ssl_object'
_A0='access-denied'
_z='/ietf-sztp-bootstrap-server:get-bootstrapping-data'
_y='Parent node does not exist.'
_x='Resource can not be modified.'
_w='multi-tenant'
_v='2022-05-26'
_u='2013-07-15'
_t='webhook'
_s='exited-normally'
_r='opaque'
_q='function'
_p='plugin'
_o='rpc-supported'
_n='data-missing'
_m='Unable to parse "input" document: '
_l=':device'
_k='import'
_j='application/yang-data+json'
_i='operation-not-supported'
_h='Content-Type'
_g='malformed-message'
_f=False
_e=':tenants/tenant='
_d='single-tenant'
_c='implement'
_b='callback-results'
_a='callback'
_Z='invalid-value'
_Y='unknown-element'
_X=True
_W='application'
_V='path'
_U='method'
_T='source-ip'
_S='timestamp'
_R='ietf-sztp-bootstrap-server:input'
_Q='conformance-type'
_P='namespace'
_O='revision'
_N='error-tag'
_M='error'
_L='protocol'
_K='text/plain'
_J='ietf-restconf:errors'
_I=':dynamic-callout'
_H='+'
_G='name'
_F='return-code'
_E='dynamic-callout'
_D='error-returned'
_C=None
_B='event-details'
_A='/'
import os,json,base64,pprint,asyncio,aiohttp,yangson,datetime,basicauth,urllib.parse,pkg_resources
from aiohttp import web
from pyasn1.type import univ
from pyasn1_modules import rfc5652
from passlib.hash import sha256_crypt
from pyasn1.codec.der.encoder import encode as encode_der
from pyasn1.codec.der.decoder import decode as der_decoder
from certvalidator import CertificateValidator,ValidationContext,PathBuildingError
from cryptography.hazmat.backends import default_backend
from cryptography import x509
from .yangcore import dal
from .yangcore import utils
from .yangcore.native import Read
from .yangcore.dal import DataAccessLayer
from .yangcore.rcsvr import RestconfServer
from .yangcore.handler import RouteHandler
from .  import yl
class RFC8572ViewHandler(RouteHandler):
	'\n      Implements route handler for the SZTPD RFC 8572 facade/view.\n    ';len_prefix_running=len(RestconfServer.root+'/ds/ietf-datastores:running');len_prefix_operational=len(RestconfServer.root+'/ds/ietf-datastores:operational');len_prefix_operations=len(RestconfServer.root+'/operations');id_ct_sztpConveyedInfoXML=rfc5652._buildOid(1,2,840,113549,1,9,16,1,42);id_ct_sztpConveyedInfoJSON=rfc5652._buildOid(1,2,840,113549,1,9,16,1,43);supported_media_types=_j,_A5;yl4errors={_A6:{_A7:'TBD','module':[{_G:_A8,_O:_u,_P:_A9,_Q:_k},{_G:'ietf-restconf',_O:'2017-01-26',_P:'urn:ietf:params:xml:ns:yang:ietf-restconf',_Q:_c},{_G:'ietf-netconf-acm',_O:'2018-02-14',_P:'urn:ietf:params:xml:ns:yang:ietf-netconf-acm',_Q:_k},{_G:'ietf-sztp-bootstrap-server',_O:_AA,_P:'urn:ietf:params:xml:ns:yang:ietf-sztp-bootstrap-server',_Q:_c},{_G:'ietf-yang-structure-ext',_O:'2020-06-22',_P:'urn:ietf:params:xml:ns:yang:ietf-yang-structure-ext',_Q:_c},{_G:'ietf-ztp-types',_O:_v,_P:'urn:ietf:params:xml:ns:yang:ietf-ztp-types',_Q:_c},{_G:'ietf-sztp-csr',_O:_v,_P:'urn:ietf:params:xml:ns:yang:ietf-sztp-csr',_Q:_c},{_G:'ietf-crypto-types',_O:_v,_P:'urn:ietf:params:xml:ns:yang:ietf-crypto-types',_Q:_c}]}};yl4conveyedinfo={_A6:{_A7:'TBD','module':[{_G:_A8,_O:_u,_P:_A9,_Q:_k},{_G:'ietf-inet-types',_O:_u,_P:'urn:ietf:params:xml:ns:yang:ietf-inet-types',_Q:_k},{_G:'ietf-sztp-conveyed-info',_O:_AA,_P:'urn:ietf:params:xml:ns:yang:ietf-sztp-conveyed-info',_Q:_c}]}}
	def __init__(A,dal,mode,yl,nvh):C='sztpd';A.dal=dal;A.mode=mode;A.nvh=nvh;B=pkg_resources.resource_filename(C,'yang');A.dm=yangson.DataModel(yl,[B]);A.dm4conveyedinfo=yangson.DataModel(json.dumps(A.yl4conveyedinfo),[B]);D=pkg_resources.resource_filename(C,'yang4errors');A.dm4errors=yangson.DataModel(json.dumps(A.yl4errors),[D,B])
	async def _insert_bootstrapping_log_entry(A,device_id,bootstrapping_log_entry):
		E='/bootstrapping-log';B=device_id
		if A.mode==_d:C=_A+A.dal.app_ns+_AB+B[0]+E
		elif A.mode==_w:C=_A+A.dal.app_ns+_e+B[1]+_AC+B[0]+E
		else:raise Exception('logic should never reach this branch')
		D={};D[A.dal.app_ns+_AD]=bootstrapping_log_entry;await A.dal.handle_post_opstate_request(C,D)
	async def _insert_audit_log_entry(A,tenant_name,audit_log_entry):
		B=tenant_name
		if A.mode==_d or B==_C:C=_A+A.dal.app_ns+':audit-log'
		elif A.mode==_w:C=_A+A.dal.app_ns+_e+B+'/audit-log'
		D={};D[A.dal.app_ns+_AD]=audit_log_entry;await A.dal.handle_post_opstate_request(C,D)
	async def handle_get_restconf_root(D,request):
		C=request;J=_A;F=await D._check_auth(C,J)
		if type(F)is web.Response:A=F;return A
		else:H=F
		B={};B[_S]=datetime.datetime.utcnow();B[_T]=C.remote;B[_U]=C.method;B[_V]=C.path;E,K=utils.check_http_headers(C,D.supported_media_types,accept_required=_X)
		if type(E)is web.Response:A=E;L=K;B[_F]=A.status;B[_D]=L;await D._insert_bootstrapping_log_entry(H,B);return A
		else:assert type(E)==str;G=E;assert G!=_K;I=utils.Encoding[G.rsplit(_H,1)[1]]
		A=web.Response(status=200);A.content_type=G
		if I==utils.Encoding.json:A.text='{\n    "ietf-restconf:restconf" : {\n        "data" : {},\n        "operations" : {},\n        "yang-library-version" : "2019-01-04"\n    }\n}\n'
		else:assert I==utils.Encoding.xml;A.text='<restconf xmlns="urn:ietf:params:xml:ns:yang:ietf-restconf">\n    <data/>\n    <operations/>\n    <yang-library-version>2016-06-21</yang-library-version>\n</restconf>\n'
		B[_F]=A.status;await D._insert_bootstrapping_log_entry(H,B);return A
	async def handle_get_yang_library_version(D,request):
		C=request;J=_A;F=await D._check_auth(C,J)
		if type(F)is web.Response:A=F;return A
		else:H=F
		B={};B[_S]=datetime.datetime.utcnow();B[_T]=C.remote;B[_U]=C.method;B[_V]=C.path;E,K=utils.check_http_headers(C,D.supported_media_types,accept_required=_X)
		if type(E)is web.Response:A=E;L=K;B[_F]=A.status;B[_D]=L;await D._insert_bootstrapping_log_entry(H,B);return A
		else:assert type(E)==str;G=E;assert G!=_K;I=utils.Encoding[G.rsplit(_H,1)[1]]
		A=web.Response(status=200);A.content_type=G
		if I==utils.Encoding.json:A.text='{\n  "ietf-restconf:yang-library-version" : "2019-01-04"\n}'
		else:assert I==utils.Encoding.xml;A.text='<yang-library-version xmlns="urn:ietf:params:xml:ns:yang:ietf-restconf">2019-01-04</yang-library-version>'
		B[_F]=A.status;await D._insert_bootstrapping_log_entry(H,B);return A
	async def handle_get_opstate_request(C,request):
		D=request;E=D.path[C.len_prefix_operational:];E=_A;G=await C._check_auth(D,E)
		if type(G)is web.Response:A=G;return A
		else:I=G
		B={};B[_S]=datetime.datetime.utcnow();B[_T]=D.remote;B[_U]=D.method;B[_V]=D.path;F,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_X)
		if type(F)is web.Response:A=F;M=L;B[_F]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(I,B);return A
		else:assert type(F)==str;H=F;assert H!=_K;J=utils.Encoding[H.rsplit(_H,1)[1]]
		if E=='/ietf-yang-library:yang-library'or E==_A or E=='':A=web.Response(status=200);A.content_type=_j;A.text=getattr(yl,'sbi_rfc8572')()
		else:A=web.Response(status=404);A.content_type=H;J=utils.Encoding[A.content_type.rsplit(_H,1)[1]];K=utils.gen_rc_errors(_L,_Y,error_message=_AE);N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(K,J,C.dm4errors,N);B[_D]=K
		B[_F]=A.status;await C._insert_bootstrapping_log_entry(I,B);return A
	async def handle_get_config_request(C,request):
		D=request;F=D.path[C.len_prefix_running:];G=await C._check_auth(D,F)
		if type(G)is web.Response:A=G;return A
		else:I=G
		B={};B[_S]=datetime.datetime.utcnow();B[_T]=D.remote;B[_U]=D.method;B[_V]=D.path;E,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_X)
		if type(E)is web.Response:A=E;M=L;B[_F]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(I,B);return A
		else:assert type(E)==str;H=E;assert H!=_K;J=utils.Encoding[H.rsplit(_H,1)[1]]
		if F==_A or F=='':A=web.Response(status=204)
		else:A=web.Response(status=404);A.content_type=H;J=utils.Encoding[A.content_type.rsplit(_H,1)[1]];K=utils.gen_rc_errors(_L,_Y,error_message=_AE);N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(K,J,C.dm4errors,N);B[_D]=K
		B[_F]=A.status;await C._insert_bootstrapping_log_entry(I,B);return A
	async def handle_post_config_request(C,request):
		D=request;F=D.path[C.len_prefix_running:];G=await C._check_auth(D,F)
		if type(G)is web.Response:A=G;return A
		else:J=G
		B={};B[_S]=datetime.datetime.utcnow();B[_T]=D.remote;B[_U]=D.method;B[_V]=D.path;E,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_f)
		if type(E)is web.Response:A=E;M=L;B[_F]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(J,B);return A
		else:assert type(E)==str;H=E;assert H!=_K;K=utils.Encoding[H.rsplit(_H,1)[1]]
		if F==_A or F=='':A=web.Response(status=400);I=utils.gen_rc_errors(_W,_Z,error_message=_x)
		else:A=web.Response(status=404);I=utils.gen_rc_errors(_L,_Y,error_message=_y)
		A.content_type=H;K=utils.Encoding[A.content_type.rsplit(_H,1)[1]];N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(I,K,C.dm4errors,N);B[_F]=A.status;B[_D]=I;await C._insert_bootstrapping_log_entry(J,B);return A
	async def handle_put_config_request(C,request):
		D=request;F=D.path[C.len_prefix_running:];G=await C._check_auth(D,F)
		if type(G)is web.Response:A=G;return A
		else:J=G
		B={};B[_S]=datetime.datetime.utcnow();B[_T]=D.remote;B[_U]=D.method;B[_V]=D.path;E,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_f)
		if type(E)is web.Response:A=E;M=L;B[_F]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(J,B);return A
		else:assert type(E)==str;H=E;assert H!=_K;K=utils.Encoding[H.rsplit(_H,1)[1]]
		if F==_A or F=='':A=web.Response(status=400);I=utils.gen_rc_errors(_W,_Z,error_message=_x)
		else:A=web.Response(status=404);I=utils.gen_rc_errors(_L,_Y,error_message=_y)
		A.content_type=H;K=utils.Encoding[A.content_type.rsplit(_H,1)[1]];N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(I,K,C.dm4errors,N);B[_F]=A.status;B[_D]=I;await C._insert_bootstrapping_log_entry(J,B);return A
	async def handle_delete_config_request(C,request):
		D=request;G=D.path[C.len_prefix_running:];H=await C._check_auth(D,G)
		if type(H)is web.Response:A=H;return A
		else:L=H
		B={};B[_S]=datetime.datetime.utcnow();B[_T]=D.remote;B[_U]=D.method;B[_V]=D.path;E,M=utils.check_http_headers(D,C.supported_media_types,accept_required=_f)
		if type(E)is web.Response:A=E;N=M;B[_F]=A.status;B[_D]=N;await C._insert_bootstrapping_log_entry(L,B);return A
		else:
			assert type(E)==str;I=E
			if I==_K:J=_C
			else:J=utils.Encoding[I.rsplit(_H,1)[1]]
		if G==_A or G=='':A=web.Response(status=400);F=_x;K=utils.gen_rc_errors(_W,_Z,error_message=F)
		else:A=web.Response(status=404);F=_y;K=utils.gen_rc_errors(_L,_Y,error_message=F)
		A.content_type=I
		if J is _C:A.text=F
		else:O=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(K,J,C.dm4errors,O)
		B[_F]=A.status;B[_D]=K;await C._insert_bootstrapping_log_entry(L,B);return A
	async def handle_action_request(C,request):
		D=request;F=D.path[C.len_prefix_operational:];G=await C._check_auth(D,F)
		if type(G)is web.Response:A=G;return A
		else:J=G
		B={};B[_S]=datetime.datetime.utcnow();B[_T]=D.remote;B[_U]=D.method;B[_V]=D.path;E,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_f)
		if type(E)is web.Response:A=E;M=L;B[_F]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(J,B);return A
		else:assert type(E)==str;H=E;assert H!=_K;K=utils.Encoding[H.rsplit(_H,1)[1]]
		if F==_A or F=='':A=web.Response(status=400);I=utils.gen_rc_errors(_W,_Z,error_message='Resource does not support action.')
		else:A=web.Response(status=404);I=utils.gen_rc_errors(_L,_Y,error_message=_AF)
		A.content_type=H;K=utils.Encoding[A.content_type.rsplit(_H,1)[1]];N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(I,K,C.dm4errors,N);B[_F]=A.status;B[_D]=I;await C._insert_bootstrapping_log_entry(J,B);return A
	async def handle_rpc_request(C,request):
		M='sleep';D=request;F=D.path[C.len_prefix_operations:];J=await C._check_auth(D,F)
		if type(J)is web.Response:A=J;return A
		else:E=J
		B={};B[_S]=datetime.datetime.utcnow();B[_T]=D.remote;B[_U]=D.method;B[_V]=D.path;H,N=utils.check_http_headers(D,C.supported_media_types,accept_required=_f)
		if type(H)is web.Response:A=H;O=N;B[_F]=A.status;B[_D]=O;await C._insert_bootstrapping_log_entry(E,B);return A
		else:
			assert type(H)==str;K=H
			if K==_K:L=_C
			else:L=utils.Encoding[K.rsplit(_H,1)[1]]
		if F==_z:
			async with C.nvh.fifolock(Read):
				if os.environ.get('SZTPD_INIT_MODE')and M in D.query:await asyncio.sleep(int(D.query[M]))
				A=await C._handle_get_bootstrapping_data_rpc(E,D,B);B[_F]=A.status;await C._insert_bootstrapping_log_entry(E,B);return A
		elif F==_AG:
			try:A=await C._handle_report_progress_rpc(E,D,B)
			except NotImplementedError as Q:raise NotImplementedError('is this ever called?')
			B[_F]=A.status;await C._insert_bootstrapping_log_entry(E,B);return A
		elif F==_A or F=='':A=web.Response(status=400);G=_AF;I=utils.gen_rc_errors(_W,_Z,error_message=G)
		else:A=web.Response(status=404);G='Unrecognized RPC.';I=utils.gen_rc_errors(_L,_Y,error_message=G)
		A.content_type=K
		if A.content_type==_K:A.text=G
		else:I=utils.gen_rc_errors(_L,_Z,error_message=G);P=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(I,L,C.dm4errors,P)
		B[_F]=A.status;B[_D]=I;await C._insert_bootstrapping_log_entry(E,B);return A
	async def _check_auth(A,request,data_path):
		'\n          creates/inserts audit-log entry!\n\n          returns:\n            on error: web.Response\n            success:  tuple (serial_number, tenant_name)\n        ';m='num-times-accessed';l='local-truststore-reference';k=':device-type';j='identity-certificates';i='activation-code';h='" not found for any tenant.';g='Device "';f='X-Client-Cert';V='verification';U='device-type';Q='sbi-access-stats';L='lifecycle-statistics';J='comment';I='failure';F='outcome';C=request
		def G(request,supported_media_types):
			'\n              inline func def\n            ';E=supported_media_types;D='Accept';C=request;B=web.Response(status=401)
			if D in C.headers and any((C.headers[D]==A for A in E)):B.content_type=C.headers[D]
			elif _h in C.headers and any((C.headers[_h]==A for A in E)):B.content_type=C.headers[_h]
			else:B.content_type=_K
			if B.content_type!=_K:F=utils.Encoding[B.content_type.rsplit(_H,1)[1]];G=utils.gen_rc_errors(_L,_A0);H=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(G,F,A.dm4errors,H)
			return B
		B={};B[_S]=datetime.datetime.utcnow();B[_T]=C.remote;B['source-proxies']=list(C.forwarded);B['host']=C.host;B[_U]=C.method;B[_V]=C.path;K=set();M=_C;N=C.transport.get_extra_info('peercert')
		if N is not _C:O=N['subject'][-1][0][1];K.add(O)
		elif C.headers.get(f)!=_C:n=C.headers.get(f);W=bytes(urllib.parse.unquote(n),'utf-8');M=x509.load_pem_x509_certificate(W,default_backend());o=M.subject;O=o.get_attributes_for_oid(x509.ObjectIdentifier('2.5.4.5'))[0].value;K.add(O)
		R=_C;X=_C;S=C.headers.get('AUTHORIZATION')
		if S!=_C:R,X=basicauth.decode(S);K.add(R)
		if len(K)==0:B[F]=I;B[J]='Device provided no identification credentials.';await A._insert_audit_log_entry(_C,B);return G(C,A.supported_media_types)
		if len(K)!=1:B[F]=I;B[J]='Device provided mismatched authentication credentials ('+O+' != '+R+').';await A._insert_audit_log_entry(_C,B);return G(C,A.supported_media_types)
		E=K.pop();D=_C
		if A.mode==_d:
			P=_A+A.dal.app_ns+_AB+E
			try:D=await A.dal.handle_get_opstate_request(P)
			except dal.NodeNotFound as Y:B[F]=I;B[J]=g+E+h;await A._insert_audit_log_entry(_C,B);return G(C,A.supported_media_types)
			H=_C
		elif A.mode==_w:
			try:H=await A.dal.get_tenant_name_for_global_key(_A+A.dal.app_ns+':tenants/tenant/devices/device',E)
			except dal.NodeNotFound as Y:B[F]=I;B[J]=g+E+h;await A._insert_audit_log_entry(_C,B);return G(C,A.supported_media_types)
			P=_A+A.dal.app_ns+_e+H+_AC+E;D=await A.dal.handle_get_opstate_request(P)
		assert D!=_C;assert A.dal.app_ns+_l in D;D=D[A.dal.app_ns+_l][0]
		if i in D:
			if S==_C:B[F]=I;B[J]='Activation code required but none passed for serial number '+E;await A._insert_audit_log_entry(H,B);return G(C,A.supported_media_types)
			Z=D[i];assert Z.startswith('$5$')
			if not sha256_crypt.verify(X,Z):B[F]=I;B[J]='Activation code mismatch for serial number '+E;await A._insert_audit_log_entry(H,B);return G(C,A.supported_media_types)
		else:0
		assert U in D;p=_A+A.dal.app_ns+':device-types/device-type='+D[U];a=await A.dal.handle_get_opstate_request(p)
		if j in a[A.dal.app_ns+k][0]:
			if N is _C and M is _C:B[F]=I;B[J]='Client cert required but none passed for serial number '+E;await A._insert_audit_log_entry(H,B);return G(C,A.supported_media_types)
			if N:b=C.transport.get_extra_info(_A1);assert b is not _C;c=b.getpeercert(_X)
			else:assert M is not _C;c=W
			T=a[A.dal.app_ns+k][0][j];assert V in T;assert l in T[V];d=T[V][l];q=_A+A.dal.app_ns+':truststore/certificate-bags/certificate-bag='+d['certificate-bag']+'/certificate='+d['certificate'];r=await A.dal.handle_get_config_request(q);s=r[A.dal.app_ns+':certificate'][0]['cert-data'];t=base64.b64decode(s);u,v=der_decoder(t,asn1Spec=rfc5652.ContentInfo());assert not v;w=utils.degenerate_cms_obj_to_ders(u);x=ValidationContext(trust_roots=w);y=CertificateValidator(c,validation_context=x)
			try:y._validate_path()
			except PathBuildingError as Y:B[F]=I;B[J]="Client cert for serial number '"+E+"' does not validate using trust anchors specified by device-type '"+D[U]+"'";await A._insert_audit_log_entry(H,B);return G(C,A.supported_media_types)
		B[F]='success';await A._insert_audit_log_entry(H,B);z=P+'/lifecycle-statistics';e=datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
		if D[L][Q][m]==0:D[L][Q]['first-accessed']=e
		D[L][Q]['last-accessed']=e;D[L][Q][m]+=1;await A.dal.handle_put_opstate_request(z,D[L]);return E,H
	async def _handle_get_bootstrapping_data_rpc(A,device_id,request,bootstrapping_log_entry):
		AP='ietf-sztp-bootstrap-server:output';AO='content';AN='contentType';AM=':configuration';AL='configuration-handling';AK='script';AJ='hash-value';AI='hash-algorithm';AH='address';AG='referenced-definition';AF='match-criteria';AE='matched-response';A6='post-configuration-script';A5='configuration';A4='pre-configuration-script';A3='os-version';A2='os-name';A1='trust-anchor';A0='port';z='bootstrap-server';y='ietf-sztp-conveyed-info:redirect-information';x='value';w='response-manager';p=device_id;o='image-verification';n='download-uri';m='boot-image';l='selected-response';h='onboarding-information';g='key';c='reference';Z=request;Y='ietf-sztp-conveyed-info:onboarding-information';X='redirect-information';L='response';J='managed-response';I='response-details';E='get-bootstrapping-data-event';D='conveyed-information';C=bootstrapping_log_entry;i,AQ=utils.check_http_headers(Z,A.supported_media_types,accept_required=_X)
		if type(i)is web.Response:B=i;AR=AQ;C[_F]=B.status;C[_D]=AR;return B
		else:assert type(i)==str;O=i;assert O!=_K;S=utils.Encoding[O.rsplit(_H,1)[1]]
		M=_C
		if Z.body_exists:
			AS=await Z.text();AT=utils.Encoding[Z.headers[_h].rsplit(_H,1)[1]]
			try:G=A.dm.get_schema_node(_z);M=utils.encoded_str_to_obj(AS,AT,A.dm,G)
			except Exception as a:B=web.Response(status=400);q=_m+str(a);B.content_type=O;H=utils.gen_rc_errors(_L,_g,error_message=q);G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,S,A.dm4errors,G);C[_D]=H;return B
			if not _R in M:
				B=web.Response(status=400)
				if not _R in M:q=_m+_AH
				B.content_type=O;H=utils.gen_rc_errors(_L,_g,error_message=q);G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,S,A.dm4errors,G);C[_D]=H;return B
		C[_B]={};C[_B][E]={}
		if M is _C:C[_B][E][_A2]={'no-input-passed':[_C]}
		else:C[_B][E][_A2]=M[_R]
		if A.mode==_d:Q=_A+A.dal.app_ns+':'
		else:Q=_A+A.dal.app_ns+_e+p[1]+_A
		AU=Q+'devices/device='+p[0]
		try:V=await A.dal.handle_get_config_request(AU)
		except Exception as a:B=web.Response(status=501);B.content_type=_j;H=utils.gen_rc_errors(_W,_i,error_message='Unhandled exception: '+str(a));B.text=utils.enc_rc_errors('json',H);return B
		assert V!=_C;assert A.dal.app_ns+_l in V;V=V[A.dal.app_ns+_l][0]
		if w not in V or AE not in V[w]:B=web.Response(status=404);B.content_type=O;H=utils.gen_rc_errors(_W,_n,error_message='No responses configured.');G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,S,A.dm4errors,G);C[_D]=H;C[_B][E][l]='no-responses-configured';return B
		F=_C
		for j in V[w][AE]:
			if not AF in j:F=j;break
			if M is _C:continue
			for P in j[AF]['match']:
				if P[g]not in M[_R]:break
				if'present'in P:
					if'not'in P:
						if P[g]in M[_R]:break
					elif P[g]not in M[_R]:break
				elif x in P:
					if'not'in P:
						if P[x]==M[_R][P[g]]:break
					elif P[x]!=M[_R][P[g]]:break
				else:raise NotImplementedError("Unrecognized 'match' expression.")
			else:F=j;break
		if F is _C or'none'in F[L]:
			if F is _C:C[_B][E][l]='no-match-found'
			else:C[_B][E][l]=F[_G]+" (explicit 'none')"
			B=web.Response(status=404);B.content_type=O;H=utils.gen_rc_errors(_W,_n,error_message='No matching responses configured.');G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,S,A.dm4errors,G);C[_D]=H;return B
		C[_B][E][l]=F[_G];C[_B][E][I]={J:{}}
		if D in F[L]:
			C[_B][E][I][J]={D:{}};N={}
			if _E in F[L][D]:
				C[_B][E][I][J][D]={_E:{}};assert c in F[L][D][_E];r=F[L][D][_E][c];C[_B][E][I][J][D][_E][_G]=r;R=await A.dal.handle_get_config_request(Q+'dynamic-callouts/dynamic-callout='+r);assert r==R[A.dal.app_ns+_I][0][_G];C[_B][E][I][J][D][_E][_o]=R[A.dal.app_ns+_I][0][_o];d={};d[_AI]=p[0];d[_AJ]=Z.remote;A7=Z.transport.get_extra_info(_A1)
				if A7:
					A8=A7.getpeercert(_X)
					if A8:d[_AK]=A8
				if M:d[_AL]=M
				if _a in R[A.dal.app_ns+_I][0]:
					C[_B][E][I][J][D][_E][_A3]=_a;A9=R[A.dal.app_ns+_I][0][_a][_p];AA=R[A.dal.app_ns+_I][0][_a][_q];C[_B][E][I][J][D][_E][_AM]={_p:A9,_q:AA};C[_B][E][I][J][D][_E][_b]={}
					if _r in R[A.dal.app_ns+_I][0]:AB=R[A.dal.app_ns+_I][0][_r]
					else:AB=_C
					K=_C
					try:K=A.nvh.plugins[A9][_AN][AA](d,AB)
					except Exception as a:C[_B][E][I][J][D][_E][_b][_AO]=str(a);B=web.Response(status=500);B.content_type=O;H=utils.gen_rc_errors(_W,_i,error_message='Server encountered an error while trying to generate a response: '+str(a));G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,S,A.dm4errors,G);C[_D]=H;return B
					assert K and type(K)==dict
					if _J in K:
						assert len(K[_J][_M])==1
						if any((A==K[_J][_M][0][_N]for A in(_Z,'too-big',_AP,_AQ,_AR,_AS,_Y,_AT,_g))):B=web.Response(status=400)
						elif any((A==K[_J][_M][0][_N]for A in _A0)):B=web.Response(status=403)
						elif any((A==K[_J][_M][0][_N]for A in('in-use',_AU,_AV,_AW,_n))):B=web.Response(status=409)
						elif any((A==K[_J][_M][0][_N]for A in(_AX,_AY,_AZ))):B=web.Response(status=500)
						elif any((A==K[_J][_M][0][_N]for A in _i)):B=web.Response(status=501)
						else:raise NotImplementedError(_Aa+K[_J][_M][0][_N])
						B.content_type=O;H=K;G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,S,A.dm4errors,G);C[_D]=K;C[_B][E][I][J][D][_E][_b][_s]=_Ab;return B
					else:C[_B][E][I][J][D][_E][_b][_s]='Returning conveyed information provided by callback.'
				elif _A4 in R[A.dal.app_ns+_I][0]:C[_B][E][I][J][D][_E][_A3]=_t;raise NotImplementedError('webhooks callout support pending!')
				else:raise NotImplementedError('unhandled dynamic callout type: '+str(R[A.dal.app_ns+_I][0]))
				N=K
			elif X in F[L][D]:
				C[_B][E][I][J][D]={X:{}};N[y]={};N[y][z]=[]
				if c in F[L][D][X]:
					e=F[L][D][X][c];C[_B][E][I][J][D][X]={AG:e};s=await A.dal.handle_get_config_request(Q+'conveyed-information-responses/redirect-information-response='+e)
					for AV in s[A.dal.app_ns+':redirect-information-response'][0][X][z]:
						W=await A.dal.handle_get_config_request(Q+'bootstrap-servers/bootstrap-server='+AV);W=W[A.dal.app_ns+':bootstrap-server'][0];k={};k[AH]=W[AH]
						if A0 in W:k[A0]=W[A0]
						if A1 in W:k[A1]=W[A1]
						N[y][z].append(k)
				else:raise NotImplementedError('unhandled redirect-information config type: '+str(F[L][D][X]))
			elif h in F[L][D]:
				C[_B][E][I][J][D]={};N[Y]={}
				if c in F[L][D][h]:
					e=F[L][D][h][c];C[_B][E][I][J][D][h]={AG:e};s=await A.dal.handle_get_config_request(Q+'conveyed-information-responses/onboarding-information-response='+e);T=s[A.dal.app_ns+':onboarding-information-response'][0][h]
					if m in T:
						AW=T[m];AX=await A.dal.handle_get_config_request(Q+'boot-images/boot-image='+AW);U=AX[A.dal.app_ns+':boot-image'][0];N[Y][m]={};b=N[Y][m]
						if A2 in U:b[A2]=U[A2]
						if A3 in U:b[A3]=U[A3]
						if n in U:
							b[n]=list()
							for AY in U[n]:b[n].append(AY)
						if o in U:
							b[o]=list()
							for AC in U[o]:t={};t[AI]=AC[AI];t[AJ]=AC[AJ];b[o].append(t)
					if A4 in T:AZ=T[A4];Aa=await A.dal.handle_get_config_request(Q+'scripts/pre-configuration-script='+AZ);N[Y][A4]=Aa[A.dal.app_ns+':pre-configuration-script'][0][AK]
					if A5 in T:Ab=T[A5];AD=await A.dal.handle_get_config_request(Q+'configurations/configuration='+Ab);N[Y][AL]=AD[A.dal.app_ns+AM][0][AL];N[Y][A5]=AD[A.dal.app_ns+AM][0]['config']
					if A6 in T:Ac=T[A6];Ad=await A.dal.handle_get_config_request(Q+'scripts/post-configuration-script='+Ac);N[Y][A6]=Ad[A.dal.app_ns+':post-configuration-script'][0][AK]
			else:raise NotImplementedError('unhandled conveyed-information type: '+str(F[L][D]))
		else:raise NotImplementedError('unhandled response type: '+str(F[L]))
		f=rfc5652.ContentInfo()
		if O==_j:f[AN]=A.id_ct_sztpConveyedInfoJSON;f[AO]=encode_der(json.dumps(N,indent=2),asn1Spec=univ.OctetString())
		else:assert O==_A5;f[AN]=A.id_ct_sztpConveyedInfoXML;G=A.dm4conveyedinfo.get_schema_node(_A);assert G;Ae=utils.obj_to_encoded_str(N,S,A.dm4conveyedinfo,G,strip_wrapper=_X);f[AO]=encode_der(Ae,asn1Spec=univ.OctetString())
		Af=encode_der(f,rfc5652.ContentInfo());u=base64.b64encode(Af).decode('ASCII');Ag=base64.b64decode(u);Ah=base64.b64encode(Ag).decode('ASCII');assert u==Ah;v={};v[AP]={};v[AP][D]=u;B=web.Response(status=200);B.content_type=O;G=A.dm.get_schema_node(_z);B.text=utils.obj_to_encoded_str(v,S,A.dm,G);return B
	async def _handle_report_progress_rpc(B,device_id,request,bootstrapping_log_entry):
		g='remote-port';f='webhook-results';Y='tcp-client-parameters';X='encoding';W=device_id;V='http';L=request;E='report-progress-event';C=bootstrapping_log_entry;S,h=utils.check_http_headers(L,B.supported_media_types,accept_required=_f)
		if type(S)is web.Response:A=S;i=h;C[_F]=A.status;C[_D]=i;return A
		else:assert type(S)==str;J=S
		if J!=_K:N=utils.Encoding[J.rsplit(_H,1)[1]]
		if not L.body_exists:
			M='RPC "input" node missing (required for "report-progress").';A=web.Response(status=400);A.content_type=J
			if A.content_type==_K:A.text=M
			else:F=utils.gen_rc_errors(_L,_Z,error_message=M);G=B.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(F,N,B.dm4errors,G)
			C[_D]=A.text;return A
		j=utils.Encoding[L.headers[_h].rsplit(_H,1)[1]];k=await L.text()
		try:G=B.dm.get_schema_node(_AG);O=utils.encoded_str_to_obj(k,j,B.dm,G)
		except Exception as K:A=web.Response(status=400);M=_m+str(K);A.content_type=J;F=utils.gen_rc_errors(_L,_g,error_message=M);G=B.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(F,N,B.dm4errors,G);C[_D]=F;return A
		if not _R in O:
			A=web.Response(status=400)
			if not _R in O:M=_m+_AH
			A.content_type=J;F=utils.gen_rc_errors(_L,_g,error_message=M);G=B.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(F,N,B.dm4errors,G);C[_D]=F;return A
		C[_B]={};C[_B][E]={};C[_B][E][_A2]=O[_R];C[_B][E][_E]={}
		if B.mode==_d:P=_A+B.dal.app_ns+':preferences/outbound-interactions/relay-progress-report-callout'
		else:P=_A+B.dal.app_ns+_e+W[1]+'/preferences/outbound-interactions/relay-progress-report-callout'
		try:l=await B.dal.handle_get_config_request(P)
		except Exception as K:C[_B][E][_E]['no-callout-configured']=[_C]
		else:
			T=l[B.dal.app_ns+':relay-progress-report-callout'];C[_B][E][_E][_G]=T
			if B.mode==_d:P=_A+B.dal.app_ns+':dynamic-callouts/dynamic-callout='+T
			else:P=_A+B.dal.app_ns+_e+W[1]+'/dynamic-callouts/dynamic-callout='+T
			H=await B.dal.handle_get_config_request(P);assert T==H[B.dal.app_ns+_I][0][_G];C[_B][E][_E][_o]=H[B.dal.app_ns+_I][0][_o];Q={};Q[_AI]=W[0];Q[_AJ]=L.remote;Z=L.transport.get_extra_info(_A1)
			if Z:
				a=Z.getpeercert(_X)
				if a:Q[_AK]=a
			if O:Q[_AL]=O
			if _a in H[B.dal.app_ns+_I][0]:
				C[_B][E][_E][_A3]=_a;b=H[B.dal.app_ns+_I][0][_a][_p];c=H[B.dal.app_ns+_I][0][_a][_q];C[_B][E][_E][_AM]={_p:b,_q:c};C[_B][E][_E][_b]={}
				if _r in H[B.dal.app_ns+_I][0]:d=H[B.dal.app_ns+_I][0][_r]
				else:d=_C
				D=_C
				try:D=B.nvh.plugins[b][_AN][c](Q,d)
				except Exception as K:C[_B][E][_E][_b][_AO]=str(K);A=web.Response(status=500);A.content_type=J;F=utils.gen_rc_errors(_W,_i,error_message='Server encountered an error while trying to process the progress report: '+str(K));G=B.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(F,N,B.dm4errors,G);C[_D]=F;return A
				if D:
					assert type(D)==dict;assert len(D)==1;assert _J in D;assert len(D[_J][_M])==1
					if any((A==D[_J][_M][0][_N]for A in(_Z,'too-big',_AP,_AQ,_AR,_AS,_Y,_AT,_g))):A=web.Response(status=400)
					elif any((A==D[_J][_M][0][_N]for A in _A0)):A=web.Response(status=403)
					elif any((A==D[_J][_M][0][_N]for A in('in-use',_AU,_AV,_AW,_n))):A=web.Response(status=409)
					elif any((A==D[_J][_M][0][_N]for A in(_AX,_AY,_AZ))):A=web.Response(status=500)
					elif any((A==D[_J][_M][0][_N]for A in _i)):A=web.Response(status=501)
					else:raise NotImplementedError(_Aa+D[_J][_M][0][_N])
					A.content_type=J;F=D;G=B.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(F,N,B.dm4errors,G);C[_D]=D;C[_B][E][_E][_b][_s]=_Ab;return A
				else:C[_B][E][_E][_b][_s]='Callback returned no output (normal)'
			elif _A4 in H[B.dal.app_ns+_I][0]:
				C[_B][E][_E][f]={_t:[]}
				for I in H[B.dal.app_ns+_I][0][_A4][_t]:
					R={};R[_G]=I[_G]
					if X not in I or I[X]=='json':e=rpc_input_json
					elif I[X]=='xml':e=rpc_input_xml
					if V in I:
						U='http://'+I[V][Y]['remote-address']
						if g in I[V][Y]:U+=':'+str(I[V][Y][g])
						U+='/relay-notification';R['uri']=U
						try:
							async with aiohttp.ClientSession()as m:A=await m.post(U,data=e)
						except aiohttp.client_exceptions.ClientConnectorError as K:R['connection-error']=str(K)
						else:
							R['http-status-code']=A.status
							if A.status==200:break
					else:assert'https'in I;raise NotImplementedError('https-based webhook is not supported yet.')
					C[_B][E][_E][f][_t].append(R)
			else:raise NotImplementedError('unrecognized callout type '+str(H[B.dal.app_ns+_I][0]))
		A=web.Response(status=204);return A